---------------------------------------------------------------------------------------------------
#                           Bayesian Affect Control Theory (V2.3.8)                              
---------------------------------------------------------------------------------------------------
  Author: Jesse Hoey  jhoey@bayesact.ca   bayesact.ca                                           
  Date  : April 2021                                                                            
  Copyright (C) 2017-2021  Jesse Hoey                                                           
  
  Research Team:
  Neil Budnarain, Benn McGregor
  
  Use for research purposes only.                                                               
  Do not re-distribute without written permission from the author                               
  Any commerical uses strictly forbidden.                                                       
  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          
---------------------------------------------------------------------------------------------------
                                                                                                
  See README for details                                                                        

See bayesact.ca for details and instructional videos                                           

 A C-only version of BayesACT Version 2.3
  
  see also doc/bayesactv2.pdf which is mostly the same as this, but also shows simulations
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------------------
Description
---------------------------------------------------------------------------------------------------
BayesACT Version 2 makes some substantial changes over BayesACT Version 1. It is written in C, not Python,
so its now faster - can run a 100,000 sample update in about a quarter of a second.  A python wrapper is
provided for convenience.
Other major changes are:

* The "somatic transform" is used for all connotative-denotative links. This was proposed in Neil MacKinnon and Jesse Hoey paper at the Identity conference. Its a major advance over the original BayesACT. 
* the denotative state (labels) are explicitly taken into account, so each sample is weighted by its distance from the sentiment distribution of its denotative label (using a somatic transform)
* re-identification is done on a *per sample* basis, so each agent maintains a running distribution over denotative labels. These are added as needed when the sample weights get too small, so the set of labels describing an agent grows and shrinks with the simulation (using a somatic transform).
* full variance information can be taken into account for dictionaries. I've included the recent 2015 dataset with the zip file in which the full measured covariances are computed for each identity label
* all identities are specified denotatively for a simulation and a simulation is specified using a simple language - see e.g. testsim.txt
* multiple agents can interact
* Interact can be mimicked closely without any fudging
* the planning engine is now a belief-state forward search.  Actions and observations are sampled and a belief state is propagated to each node.
  The idea is to do a fairly shallow, precise, targetted search. The search goes for precision in a targetted (by the affect control principle)
  area rather than depth in an expectet (according to the utility function) target area.
* March 2021 overhaul fixed up some remaining problems, removed SELECT_ACTION which was obsolete, made all parameters settable from the command line,
  addded CSV writing, and added the computation of Interact optimal behaviours and re-identifications as remedys for sampling breakdowns.
  Also see below under "parameters" for more updates and usage notes on the new V2.1 edition March 2021
  Also: removed mimic_interact as this can be simulated with parameter settings (choosing lots of samples and undispersed distributions so small alpha and gamma)

// THIS DOES NOT WORK YET -- you have to take turns still 
* both agents can act simultaneously (turn dynamics are fully implemented, and there is no "turn" variable anymore)


---------------------------------------------------------------------------------------------------
Installation
---------------------------------------------------------------------------------------------------
Requires the following:
* gcc - on mac you can download the XCode package 
* GSL - get it from  www.gnu.org/software/gsl

**** USE GSL VERSION 2 if possible. If not possible, then you must edit the fvarsample.h file to flag this - see below

To install, open a terminal, find the file and 
>unzip bayesact2.1.zip
>cd bayesact2.1/source/
>make
**** if using GSL version < 2, edit fvarsample.h and uncomment GSL_VERSION_2 first


// there is also testfvarsample.c and testbayesact.c but these have not been tested
To use:
> ./bayesactsim -h
displays a usage message

> ./bayesactsim ../examples/testsim4.txt

where (optional) num specifies the number of samples to use and testsim.txt is the
simulation specification file to read from 

To use emotions, specify a -e flag after the function call
> ./bayesactsim -e testsim.txt

-e flag is overruled by -m

---------------------------------------------------------------------------------------------------
Notes on the code
---------------------------------------------------------------------------------------------------
The basic structures are found in fvarsample.h, the major functions in fvarsample.c
There are four major components -- see also the diagram in doc/bayesactv2.pdf

* agents represent agents (!) and are agentState objects which have a name,
  filenames for transient impression and emotion dynamics parameters,
  a pointers to dictionaries for identities, behaviours and modifiers

* dictionaries represent sentiment dictionaries. These are not replicated, so more than one agent
  can point to the same dictionary. they are used by the somaticTransform to map bertween
  connotative and denotative state.  Dictionaries can be of four types (enum dict_type)

  * MEAN - a label followed by SSD double
  * SD - a label followed by SSD doubles (the mean), followed by SSD doubles (the std. devs. on the mean)
  * COV  - a label followed by SSD doubles (the mean), followed by SSD^2 doubles (the covariances on the mean)
  * SAMPLE - (raw set of samples - not implemented yet)
  * GAUSS_MIX - (a label followed a number M  - number of mixtures, followed by M sets of weights + COV or SD parameters - not implemented yet)

  gender is handled by simply specifying a dictionary that has the correct sentiments for that gender.
  so the old interact dictionaries (e.g. fidentities.dat) which had male and then female means, the first
  three are used (male only). To use these for female identities, they'd have to be split into two dictionaries
  
* dyads represent an interaction, and are dyadIndex objects which have an index for the dyad,
  and indices for agent and client in the interaction

* samplerState represents the parameters and samples for each interactions. It has two arrays of size
  __CURRENT_NUM_INTERACTIONS which hold parameters and samples for each interaction. These arrays
  are accessed using a single integer (as given by a dyadIndex structure). The arrays hold:

  * the parameters are sentimentParams which includes all the covariance matrices for the updates
    and the parameters of impression formation equations (M) and emotion equations (R,E,Q)

  * the samples are an array of sample structures which have 2 double arrays (for f and tau), and
    then integer indices for agent id, client id, agent behaviour id, client behaviour id and finally a double weight

- there are a number of enums and other constants defined - see fvarsample.h for details. Of note:
  * SSD is the dimensionality of the emotion space (3 by default: EPA)
  * NGT is the number of grammar terms - in case we want to add settings in later


---------------------------------------------------------------------------------------------------
Python3 wrapper
---------------------------------------------------------------------------------------------------

To use the Python3 wrapper, you have access to only the very basic functions right now:
- to initialize, you provide the same arguments as to testbayesact in a python3 call:
   initializeState(numsamples, num_plan_samples, timeout, simfile)
- to get the next action and emotion for the agent for dyad i:
    (agent_action,agent_emotion) =  getNextActionEmotion(i)  
- to update the state with the xx observations, the client action and emotion and agent action and emotion as specified (all strings) for dyad i
  returns the number of nodes in the plan tree - if < 0 this means this failed - something was not in a dictionary (probably an emotion label)
    numnodes = updateState(i,xx_obs,client_play,client_emotion,self.agent_play,self.agent_emotion)


---------------------------------------------------------------------------------------------------
Usage notes
---------------------------------------------------------------------------------------------------
./bayesactsim -h
gives the options

example simulation would be (from source/) (tstout.csv is optional)
./bayesactsim -o ../output/tstout.csv ../examples/testsim7.txt 

There is also a program output/csvtolatex.py to convert the csv to a latex table by including those columns that you want for
each row (third argument) and second row (third argument). The reason you have to specify columns for each of two turns is that
the AGENT in each turn is different, so that needs to be reversed for the second turn when showing in a table. However, this
program assumes the turns are sequential (no repeated turns), and this could be fixed.

python3 csvtolatex.py ./testout4f.csv ./colstolatex ./colstolatexi  >& coalminer2.tex
then pdflatex coalminer.tex which includes coalminer2.tex (or whatever you called it)

See the simulation file in examples/testsim7.txt for an example that can be used.
That file contains the following elements:

* A set of agents (A,B,C ...) are defined, complete with parameter as follows for each agent:
  * persona  (denotative distribution over identities held roughly as "self" for this agent)
  * identity, behaviour and emotion dictionaries (files)
  * impression formation equation file
  * emotion formation equation file
  * alpha  (variance in the affect control principle)
  * beta (usually 1/10th of alpha, variance in the connotative transition function)
  * delta (variance in the denotative transition function)
  * gamma (variance in the somatic transform, specified as
       (1)  in the dictionary as a variance (SD) or covariance (COV);
       (2) hard-coded and uniform as DTYPE_MEAN_VARIANCE in fvarsample.h or
       (3) provided on the comman line with argument -ga (MEAN)
       NOTE: I decided to make gamma a property of a dictionary, as that is really how it is meant to work.
       In order to "override" this value, you have to write out a new dictionary (if you want to override on a per-identity basis),
       or you have to override the entire dictionary by labeling it MEAN and using DTYPE_MEAN_VARIANCE or -ga on the input line.
  * number of samples

NOTE: The data that will be used, including all impression formation equations are read from inputparser.h: DICT_PATH

* A set of dyads (A-B, A-C, ... , B-A, B-C, ...) are specified (Assume: A cannot act on B without B being acted upon so the mirror dyad should always exist, but if it does not, the simulation can still be carried out (as in Interact), but only from a single agent's perspective. The following is specified for each dyad:
  * institutions (list of names in the set) - only if the dictionaries contain institutions
  * random seed (optional)
  * agent identity (denotative distribution over identities held roughly as "self" for this agent)
  * client identity (denotative distribution over identities held roughly as "self" for this agent)

* A specification of how the simulation is to be carried out (interactively or batch)
* An events file if batch - see ../examples/tests4fn.events  that specifies for each event
  * agent and client actions (can be *=optimal, !=interact, blank=none, or a behaviour_label, and can have a following + indicating noise is added according to DISRUPT_NOISE in fvarsample.h
  * agent and client emotions (can be *=optimal or an emotion label), ignored if -e flag not specified
* If running interactive,the simulation proceeds by 
  * asking which dyad you want to simulate - type the index or "enter" to use the same as last time
  * printing out summary statistics for this dyad, including
    * average sentiments (fundamentals and transients)
    * average labels for those sentiments (distribution over denotative states)
    * deflections (expected)
    * deflections (from average state)
    * emotions and labels
  * printing out best next actions for agent    
    * best next actions take (expectation over denotative labels)
    * best next action to take (denotative label based on average state)
  * asking for the agent's action (if he can act - i.e. if the "turn" allows for it). "enter" here means
    that agent does not act - all samples that have turn=agent are discarded
    if query : bayesact is specified in the simulation input file, then the default action is taken
  * printing out best next actions for client (see (3))
  * asking for the client's action (if he can act - i.e. if the "turn" allows for it). "enter" here means
   that client does not act - all samples that have turn=client are discarded
   if query : bayesact is specified in the simulation input file, then the default action is taken
  * actually doing the update  - you have to select the same dyad again to see what happened
  *  go to (1)

SAMPLING
--------
- num_samples (specified on the command line with -n or when creating an agentGroup in testbayesact.py): number of samples to use
  in the sampler. More samples means better simulation, but slower.
- default_action (specified on the command line with -d) : if set, then getNextAction will compute the average next state (like in ACT)
  and then somatic transform that f_b value to get the next action. If not set, then getNextAction computes the full somatic transform,
  yielding the expected distribution over the denotative action space. 

REIDENTIFICATION
----------------
- RESET_THRESHOLD (fvarsample.h) : governs how likely a re-identification in the denotative space will occur. This threshold is
  on the particle filter weights (PER SAMPLE). If the weight of a sample falls below this threshold, its denotative identity value
  will be re-chosen based on its sentiment value. This will happen first for CLIENT identity, and if that doesn't work, for AGENT
  identity, and if that doesn't work, for both. If that doesn't work, then the sample is discarded. If all samples are discarded
  then the old samples are simply copied over, and the roughening noise is doubled (a bunch of additional noise is thrown in).



---------------------------------------------------------------------------------------------------
Older notes, to-do lists etc  some material in what follows may be out of date
---------------------------------------------------------------------------------------------------


// 2.3.8 notes 
gamma is the variance in the dictionary -- this can be specified IN the dictionary (if type SD or COV)
but if the dictionary type is MEAN, then its DTYPE_MEAN_VARIANCE *unless* -ga is specified on the command line

to run an "interact"-like simulation
use testsim9.txt (basic version) and testsim9.events (so one agent only runs BayesACT the other does interact actions)
set the following in fvarsample.h

#define DO_DENOTATIVE_IDENTITY_UPDATES 0
#define NEVER_RESET
#define DO_INTERACT 

and run with

bayesactsim -ga 0.1 ../examples/testsim9.txt

to get a nice version with a bit less chaos, filter like this:
./bayesactsim ../examples/testsim4f.txt |  grep --after-context=15 'fundamental output average\|perceives\|deflection' | grep --after-context=1 'identity\|perceives\|deflection:'


to run the equivalent "full bayesact" simulation, use testsim9.txt and set -ga 0.05
./bayesactsim -ga 0.05 ../examples/testsim9p.txt

Because of the total derivative, a full bayesact simulation may give rise to behviours that are not the same as the interact agent, even if everything else is held the same.
This is because BayesACT solves for the optimal distribution of the full 9D fundamental sentiment in the posterior, and then computes marginals over behaviours. Thus, behaviours selected by BayesACT as "optimal" may "require" a re-identification of the sample as well to a new identity. Should that not be possible, the resulting behaviour may not be different. For example, the nurse who pretended to be a secretary was fine doing a behaviour like "dress" but this required her to think of herself as a secretary. If she is forced to remain a nurse, she would select a different identity (like "confide in" I think - check).

So, to replicate *Exactly* an interact simulation, you can use testsim7.txt which pits a student and a tutor who share a complete definition of the situation. Further, you must use testsim7.events where only bayesact actions are taken, and you must set
all alphas: 0.001 (in testsim7.txt)
gamma: 0.001 (on command line, see below)
in fvarsample.h:
#define RESET_THRESHOLD 1e-307
#define NEVER_RESET
#define DO_INTERACT 
#define DO_DENOTATIVE_IDENTITY_UPDATES 0
#define DEFAULT_AGENT_DELTA 0.1   // this is in bayesactAPP.c and really doesn't matter since DO_DENOTATIVE_IDENTITY_UPDATES is 0

then you get exactly what interact produces (or at least very close) 
./bayesactsim -ga 0.001 ../examples/testsim7.txt
OR check testout.csv after running:
./bayesactsim -ga 0.001 -o ../output/tstout.csv ../examples/testsim7.txt

You can also specify "!" as actions in the "events" file and then you get exactly interact (the partial derivatives)


-------------------
2.3.8 TODOs (see also more below)
---------------------
still to verify that it is working properly but have no reason to believe it is not
* fix identity specification for each interaction so it can be pairs of identies (not factored over agents)
* emotions
* pywrapper.py check/test/update
* check xx state
* planning (forvalue.c)
* update all documentation to reflect new codebase

new to-do items:

* integrate with a generic POMDP solver so the BayesACT overlay can be "attached" to any problem
* multithreading/neural implementation of impression formation and posterior calculation
* auto-fill of mirror dyads
* 



// 2.3.7 notes April 2021 notes (to be inserted properly into the documentation once I get to that

new APP files:

defaultAPP : need to have DO_DENOTATIVE_IDENTITY_UPDATES = 0 for the basic institution-free version

defaultAPP has x_a' being any identity at all

institutionalAPP: need to have DO_DENOTATIVE_IDENTITY_UPDATES to take institutions into account

institutionAPP has x_a' be any identity from the set of institutionally correct ones (for this dyad)

bayesactAPP: need to have DO_DENOTATIVE_IDENTITY_UPDATES to take institutions into account

bayesactAPP is a full bayesACT simulation for denotative identity --- that is the x_a' is something close to x_a, but institutionally correct, and is weighted by its distance from f

Institutions are applied to identities and behaviours, not to emotions
If any sentiment set (Dictionary) does not contain any label for an institution, and that institution is selected,
the value that will be returned by the somatic transform (DENOTATIVE) which is the best across all identities, regardless
of institution
See the extensive note in buildInstitutionFilterList() in inputparser.c for more information on this issue

./bayesactsim -ga 0.1 -aa 0.01 -ab 0.01 -ac 0.01 -o ../output/tstout.csv ../examples/testsim9.txt | grep --after-context=15 'fundamental output average\|perceives' | grep --after-context=1 'identity\|perceives'

To do a simulation of pure interact, then do the following
1) uncomment DO_INTERACT in fvarsample.h
2) specify all "!" for the acting agent in an events file (see testsim9ii.events)
3) note that the two BayesACT agents will still run, but will be "forced" to take the
   interact optimal action for the original identity setting (first on each line for each dyad)
   thus their interpretation of the situation may differ, but the output file should show the
   interact simulation that should be exactly the same as an interact simulation. IF it doesn't, let me know





-----------------------------------------------------------------------------------------
updates to 2.2
this version can do a full interact simulation simultaneously so you can "verify" your parameters.
Synopsis is that if you can run alongside interact and track it fairly well *under optimal actions only*,
then you can be somewhat confident in your parameter choices. A full readout of the Interact simulation
is shown on the standard output (terminal window), but the CSV file (if output) will also contain some
indicators you can track. The last five columns of the CSV file give the
(1) indicator (0/1) of whether the BayesACT optimal agent action suggested (label) is the same as the Interact one
(2) Euclidean distance from BayesACT optimal agent action (EPA) to Interact one
(3) indicator (0/1) for client action
(4) Euclidean distance for client action
(5) average absolute difference between BayesACT transient and interact transient
Basically you want (1) and (3) to be 1 and (2,4,5) to be close to 0

The six "resets" columns that precede that show how many resets are done. You can "force" this to be 0 by setting RESET_THRESHOLD in 0.0 *and* defining NEVER_RESET in fvarsample.h, in which case only the "hard" resets will be shown in the "hard resets" column.You may want resets or not, but these columns can show you what is going on a bit with that.

This version can also handle disruptive events now  - so if one party does something disruptive, but the other party continues to act according to their original identity, then the pair will restore the original identity set. If the other party re-identifies, then the pair may assume a different identity. In BayesACT, the re-identification of both identities and behaviours is necessary, so you have to play with the RESET_THRESHOLD (in fvarsample.h) in order to get this right.

the example in testsim5.txt is for a simple tutor-student interaction and if you try with testsim5.events (all optimal), you get a match with interact at the end of 8 iterations of about 0.124, which is close (so each element of the transient is off by about 0.1).
the csv output for this is in output/tstout-tutorstudent-opt.csv-- u
sing ./bayesactsim -n 1000 -ga 0.05 -aa 0.05 -ab 0.05 -ac 0.05 -o ../output/tstout-tutorstudent-opt.csv ../examples/testsim5.txt

If you try with testsim5d.events though (this is where there is a disruptive event, followed by optimal events for tutor and student), you get 0.31 so its much bigger,
but in the end the identities still match (tutor/student) even though the pair has gone through some alternative identities during the simulation
the csv output for this is in output/tstout-tutorstudent-disrupt.csv

If you try with testsim5e.events (this has a disruptive event followed by the optimal BayesACT action for whatever identities BayesACT currently re-identifies as), you getthe same transient (0.05 difference) because BayesACT is attempting to make the two consistent and has re-identified (As a "gun moll" and a "grind") and is taking actions consistent. The interact simulation does *not* re-identify, but it obtaining roughly the same transients (at much higher deflection). The deflection for Interact at the end of this simulation is 11, while for BayesACT (who has re-identified) it is 0.5
the csv output for this is in output/tstout-tutorstudent-bayesactopt.csv
----------------------------------------------------------------------------------------


A refined simulator is in bayesactsim.c.  This one makes the following additions:

- it will update multiple dyads simultaneously. Thus, if "dave" does behaviour B to "tom", this will be
interpreted in the dave-tom dyad and in the tom-dave dyad if both exist. B is interpreted by dave in the
dave-tom dyad and by tom in the tom-dave dyad. If the tom-dave dyad doesn't exist in the simulation file,
then its ony interpreted by dave in the dave-tom dyad.

- Optimal behaviours are calculated for each *agent* now, rather than for the principal (first) actor in a dyad.
Recall that a dyad has two actors, and the first is the one from whose perspective the dyad is being interpreted. Thus
interaction: dave : tom
means that this dyad is dave's interpretation of the events
If the tom-dave dyad (tom's perspective) is not specified in the simulation file, then the optimal actions for tom will be computed from dave's
perspective. If the tom-dave dyad is specified, then the optimal actions for tom will be computed from tom's perspective.

- it can read in "events" files that specify sequences of events and it will run them automatically.
for an example of this, see the simulation file person1.txt, which specifies
events: person1.events
instead of
num_terations: 50
This file has lines of the form
dave : A1 : E1 : tom : A2 : E2 :
which specifies that dave does A1 and expresses emotion E1 to tom, while tom does A2 and expresses E2 back to dave (on the same turn)
if A1 is * then the optimal action/emotion is taken (optimal is computed using averaged, expectation or plantree methods)
if A1 is missing, then it means dave doesn't act in this turn
if E1 is either * or missing, it means the default emotion is expressed by that agent   


---------------------------------------------------------------------------------------------------
Parameters
---------------------------------------------------------------------------------------------------
There are six major parameters that can be tuned for different behaviours.
The effect of these parameters on any given simulation is only somewhat understood.

NON-MEASURABLE parameters governing the sentiment dynamics
----------------------------------------------------------
- \alpha and \beta  (set in testbayesact.c) : these are the main BayesACT parameters. \alpha governs
  the strength of the affect control principle, while \beta governs the strength of the sentiment inertia
  \alpha=\{\alpha_a, \alpha_b, \alpha_c \} where \alpha_i for i\in\{a,b,c\} corresponds to 1/w_i where w_i
  is the deflection "weight" in ACT. This was likely introduced in ACT to handle cases where different aspects
  of the deflection would be more or less important in the process of behaviour selection
  e.g. deflection in power is more significant than deflection in activity, and one would prefer to gain a reduction of 1 in
  power deflection to a reduction in 2 in activity deflection.
  It could also be used to model different importance of different entities
  - e.g. deflection of self identity is more  important than deflection of other identity
  So, differences in the \alpha_i parameters for different i will govern these aspects
  However, in BayesACT, the relationship between \alpha and \beta is also important
  \beta governs the strength of the intertial term - how much we expect sentiments to stay the same over time
  in a particular interaction
  \beta=\{\beta,\beta_c\} since behaviour sentiments are not "carried over" through time.
  if \alpha >> \beta, then the agent will be much firmer in his beliefs about his identity remaining stable over time,
  and will tend to ignore events that challenge this identity more
  if \beta >> \alpha then the agent will be more likely to "re-invent" himself in each timestep to suit the ongoing interaction

  \alpha and \beta are not measured right now, but I imagine they could be

Where to find and set paramters in bayesactsim
1) For most users, on the command line will do fine. You can ask for help with -h and see the parameters that can be set: alpha, beta, and gamma
2) For those who want to get their hands dirty,
    a) the gamma parameter is specified in the dictionary file you specify in the input files e.g. testsim.txt  - but if MEAN you can specify it on the command line, or it will be set to DTYPE_MEAN_VARIANCE as usual
    b) the alpha parameter is passed into initializeBayesACT but really should be a property of the tydnamics-X.file
    c) the delta parameter is a property of the subclass of pdAPP.c that you write for your specific applications
    d) the beta parameter is lumped in with alpha in my recent papers

So these three parameters go a long way to controlling what happens in your simulation. Primarily they determine what kind of "Resetting" is going on. This can also be controlled by the the RESET_THRESHOLD which is defined in fvarsample.h. This is fixed at 1e-20, but it can be lowered or raised in the code (followed by a re-compile). The reason this is not an input parameter is that it controls at which point the sampler is considered to have "broken" (to have zero weight). Essentially this controls whether the agent being simulated has a hair trigger when it comes to re-identification or not. If you set RESET_THRESHOLD lower (smaller), then you get a more robust agent who is able to maintain his identity in the face of highly deflecting events. If RESET_THRESHOLD is set very large (closer to 1), then the agent will be more likely to re-identify the client or itself.
However, I believe that setting it at a fixed, small number, which I have chosen arbitrarily at 1e-20, we can be fairly certain that we are simulating the same class of people, and that other differences are due to modifications of the alpha/delta/beta/gamma parameters above.
To sum up, there is an overall scale factor involved in this operation and it can be controlled with the overall scale of the parameters alpha/beta/gamma/delta. The RESET_THRESHOLD gives a mechanism for specifying this overall scale, but it should not be relevant for the computation as a whole as it will only set the relative values of things.

Basic usage can be controlled on the command line with alpha/beta/gamma being specified, with gamma specified in the dictionary files referenced in yoru simulation input file (e.g. testsim.txt), or on the command line (or as DTYPE_MEAN_VARIANCE in fvarsample.h if all else fails). 

In BayesACT there is a quite a bit going on "under the hood", and you, the user of the BayesACT software has the responsibility to understand what is going on "under the hood" so you can better interpret your simulations.

One key element that is "under the hood" is the somatic transform.
The entire operation works as follows (this is in bayesactsim.c): 

1) sample from the posterior over fundamental sentiments, using a deterministic transient (still not sure why I did this and what effect it has)
2) sample from the denotative state  -- this is where institutional filters come in if there are any 
3) weight by denotative-connotative by somatic transform and by the evidence, which is uncertain
4) sample an action to take, possibly after looking forward in time and projecting expected value backwards
5) if weight=0 - the sampler "breaks" and we have to re-identify someone, so start with re-identifying the client, weight behaviour by evidence
6) then if that doesn't work, re-identify both and set the behaviour to be the evidence - this essentially sets the event to be whatever the Interact computation would be with both actors re-identified to be as consistent with the observed action as possible. 
7) if that doesn't work, take the default action (the state remains the same, all evidence is ignored - essentially this event is skipped) or exit.

MEASURABLE parameters governing sentiment dynamics
--------------------------------------------------
- covariances in sentiment dictionaries are very important (they replace what used to be called \gamma in the original statement
  BayesACT. If dictionaries don't have variance information, then the parameter DTYPE_MEAN_VARIANCE in fvarsample.h specifies a
  single generic variance for all sentiments in that dictionary. This is the same as \gamma.


PLANNING
--------
- TOLERANCE and DISCOUNT (fovalue.h) :  Used when planning only. Really DISCOUNT is a property of the domain of application,
  so here I'll focus on TOLERANCE only. TOLERANCE governs how deep the plan tree will be grown - the max depth of the plan tree
  the tree will stop expanding when DISCOUNT^depth > TOLERANCE
  so max depth = log(TOLERANCE)/log(DISCOUNT)
- num_plan_samples (specified on the command line with -p or when creating an agentGroup in testbayesact.py): number of samples to use
  in the planner. More samples means better simulation, but slower. If not specified or zero, then planning is not attempted. This
  is a test code thing though - if you don't want to plan, you just don't call seekNextAction (use getNextAction instead), and don't
  call pruneTree after an update
- timeout (specified on the command line with -t) : how long to plan for. Longer planning times will be slower (duh) but will result
  in actions that better optimize individual reward for the agent (while taking into account the emotional alignment).
- emotion_planning (specified on the command line with -ep) : whether to sample from the distribution over current emotions when building
  the plan tree. The default is to simply
  1) if emotions are being used at all (if -e is used): use the default emotion
  (right now this is the denotative emotion that has the highest likelihood, but this could be the emotion computed from the current average state)
  2) if not: a null (dummy) emotion. In this last case, the plan tree really doesn't include emotions, and we could define a version that doesn't
  even add them into the tree, but the overhead this would reduce the tree by is minimal (one extra pointer per node only).


---------------------------------------------------------------------------------------------------
-----------------------------      TODO list   ----------------------------------------------------
---------------------------------------------------------------------------------------------------

NEW MARCH 2021

  * Planning method is still left unchecked - do not use ae_method=PLANTREE. 


  * initialization of samples - should be in the APP files not hardcoded in getSampleSet() in fvarsample.c

  

  * optimize the nearest neighbour function (in DENOTATIVE somaticTransform)


  * BayesACT self - replaces "persona" agents have situational and fundamental self-sentiments and their initial identities
    are drawn from these, and these are updated after each event, and events contribute back to them

  * understand how we can trade-off the variance of emotion (which is measured) vs. alpha & beta (Which are explicitly set).
    However, we could measure beta as well I suppose ... and alpha. The dynamics equations are estimated based on some regressions,
    so must have some variances that we could measure. 

  * make simple generic batch ACT code simulation in python  -- this is testbayesact.py, a work in progress. Right now it includes
    only the basic functionality. Planning is not included yet - this would have to be added to pywrapper.c

  * TODOs in forvalue.c

  * pruning the planTree - right now it doesn't work very well because of the large denotative emotion space (so you rarely re-use
    an emotion observation branch. Two ideas :
         * prune in EPA space - compare EPA(client behaviou/emotion observation) with EPA(client_behaviour/emotion in tree branch)
	 * prune using the denotative *state* - given observation, do belief update, extract x'_bc, x'_x and prune based on these

 
  * figure out how to link the files so it makes sense - right now you have to change the #include
    in fvarsample.c and then change the Makefile to be the new xxAPP.c xxAPP.h whcih is awkward
    *** the RIGHT way to do this is the have a single defaultAPP.c class which *reads* a POMDP specification from a text file
    in e.g. SPUDD format.  Then, the user has to specify a simulation file (testsim.txt), in which are specified for each interactant,
    a set of dictionaries, and a SPUDD MDP file!

  * investigate cosine distance as a replacement for euclidean distance. All my students are using this because it works better
    in spaces that are not isotropic

  * specify no emotion option direclty in the testsim file - if there is no "EMOTION" dictionary specified, then the "no_emotion" flag is set
    **for that agent only** -- wait then it complicates things because one agent doesn't use emotions at all .. could work in fact
    could simulate an emotionless agent interacting with emotional agents. Is it a property of agents or dyads? - I think agents. So
    an emotionless agent would ignore signals from other agents and not produce any signals (or maybe always send a "no emotion" signal?)
    

  * implement SAMPLE dictionaries and somatic transform
  * implement GAUSS_MIX dictionaries and somatic transform
  * work out how to add/remove Gaussian mixture components to add/delete state (learn state structure)
  * implement traits (e.g. angry mother, dissapointed lawyer)
  * handle gender a bit better somehow - make a property of agent?
    Right now its fixed by specifying a dictionary for each agent that is appropriate
  * document the code better


  ------------------- TODOs DONE or IGNORED
  * organize the code structure and make release packages
  * do not use mirror dyads -- must be fixed with CSV printing above 

  * action filters - institutional filters on behaviours in setNextSample -- must be in the xAPP.c files in fact that these are applied

  * add settings (institutions). Basically each sample will be augmented with a new connotative f_s,tau_s and denotative x_sf
    so NGT=4 and NGTD=5, add SETTING to c_ and d_entity_type, update NET_C and NET_D, agent_map and client_map
    we can also add some observations of the settings
    then, in a somatic transform from sentiment->x, we can "filter" the possible denotative identities we test over -
    this implies passing in another argument (the setting) and then checking each dictionary entry (which must then have a
    set of flags for each setting indicating which it applies to) against the proposed setting, and using it only if it is relevant.
  * implement institutional filters

  * For now, the CSV output works only with a single dyad specified in the input file (A single "interaction") -- e.g. testsim5.txt and does not work with mirror dyads turned on. I have to work out how to better represent the header for the CSV file - may have to be updated on the first pass through the main loop, and then the file needs to be saved, closed, and then re-opened and read and then rewritten with the header. --- I did this and fixed it all up - but its all fixed headers so you can't change anything 
  * write example code for doing multiple simulations simultaneously - i.e. when sally acts on bob,
    both the sally-bob and bob-sally dyads should be updated
    --- this is now being done with the events idea
  * write a document describing how BayesACT V2 works in more detail 
  * replace euclidean_distance with weighted distance in DENOTATIVE somatic transform -
    and then replace the min distance computation with drawing a sample from those probabilities
    see TODOs in somaticTransform() in fvarsample.c - DONE, but it slowed things down a LOT so was removed
  * make it so the planning engine will completely ignore emotions if they are not being used. - each emotion branch could be "no emotion"
    to make it consistent, but the state updates then have to use no emotion as well - I think this is done by simply passing in NULL
    for agent and client emotion  **** DONE 
    ALSO: need to make it so "no emotion" is located in the dictionary automatically - now its hardcoded in getNullEmotion  - see below -
    ALSO: right now its weird because all the updates in the "no emotion" case happen with emotion=NULL, but in the plan tree its represented as "no emotion"
          but "no emotion" is really an emotion with an EPA signature (which is ignored because we check for use_emotions and
	  call sampleNextState and computeWeight with the emotion set to NULL .. so this is somewhat awkward and could be made ... better
	  ** possibly: each dictionary adds a single EMOTION_NULL value at 0,0,0 or something - add this at INDEX=0?
    I implemented a version where I would create an artificial "NULL_EMOTION" entry in the emotions dictionary, but that in the end made no sense since
    it made an artifical emotion sentiment that could be used anywhere. This was committed as version 788, but I rolled back to 787 and undid all that.
    leaving this for now. To ignore emotions, you simply don't set the -e flag and all emotions in the plan tree are index 0 which is whatever it is in
    the dictionary, but it doesn't actually matter (does not play a role at all)
  * fix the plan tree observation probabilities - it seems there is no probability distribution over the client actions - why not? Find this
    and fix it! --- DONE - I was doing this totally wrong. Now its fixed so it keeps a count of samples and normalizes to get the probabilities
  * add roughening noise -- DONE
  * add weights for emotions (both for agent and client)  -- DONE
  * handle the full reset case - when all weights go to zero right now it simply memcpys the
    old samples exactly. Either return to original state or some generic state? -- LEFT AS IS EXCEPT
    INCREASE ROUGHENING NOISE
  * update agent_type so its called reset_type and add a new "entity_type" enum (See fvarsample.h)
    which will allow us to easily connect x with f  -- see entity_type and CHESS in fvarsample.h --- DONE
  * make x in sample a vector of size NGT - so there is a correspondence directly with f and
    the somatic transform can be applied in a general way across all parts of X in sampleWeightNextSentiment -- DONE
  * change dictionary in agent so its an array of NGT dictionaries - then we can iterated over it when calling
    the somaticTransform in sampleWeightNextSentiment -- DONE
  * change dictionary specification in testsim and parser - see testsim.txt-- DONE
  * update getSampleSet so it works for any NGT size  (not just agent and client ids) -- but this also requires the array of dictionaries
    I need to make this all consistent. --- DECIDED NOT TO DO THIS
  * fix the stupid indices - its all confused with turn, agent index and reset_type entity_type etc
     basically  I need to have a seperate turn_type which is just AGENT and CLIENT,  <<--- decided not to do this either
  * change turn to also be a somatic transform -- DONE BUT SEE BELOW - turn was removed
  * implement TURN as entity type and turn dictionaries with two entries - "take turn" and "give turn" where "take turn" is more powerful
    and active than "give turn" - e.g. defer to,-0.15,0.45,-0.44 and "deny" -0.89,1.84,0.21 or maybe (from Indiana dictionaries):
    close to [0.5,-0.5,-0.5] : "follow" 0.38,-0.22,-0.70 "yeild to" 0.27  -0.24  -0.15,  or wait on,    0.49  -0.06  -0.57,
    close to [0.5,0.5,0.5] : "cue",    0.43   0.61   0.50,  "direct"    0.44   0.69   0.66,  
    DONE -- but see below, turn was removed
  * implement a full sample reset instead of the per-sample version - basically just avoid re-setting individual samples and then
    go through all samples if the total weight is zero and do a reset for client, for agent, and then for both. <-- DID NOT DO THIS
    BECAUSE IN EFFECT, WHEN THE TOTALWEIGHT->0, ALL SAMPLES HAVE BEEN RE-IDENTIFIED ANYWAYS, SO THE ONLY THING LEFT TO DO IS TO COPY
    OVER THE PREVIOUS SAMPLES WITH ADDED NOISE
  * remove turn -- DONE! -- see sampleWeightNextSentimentWithEmotion essentially what happens is that if one agent acted, then
    the dynamics for that agent are used (e.g. agent or client dynamics), but if both acted, then a random choice is made. What this
    does is allocate samples randomly to each action.  The key is that the *same* sample can be used arbitrarily for different or same
    turns on each time step, it doesn't actually matter. 
  * implement external state X --- I think this will mean adding something to d_entity_type (e.g. XSTATE) and then changing the
    dynamics of X in the updating functions - call a function updateXSample in getNextAction and in sampleWeightNext...  DONE
  * add in XX observation -- DONE
  * implement POMCP planning and external state (X) -- started work on this - first attempt using XX as part of sample state seemed
    really ugly so I've commented it out everywhere with the code XX - DONE - see forvalue.c/.h
  * remove sstate->vector_dim and related, use VD everywhere instead
    

---------------------------------------------------------------------------------------------------
File Listing:
---------------------------------------------------------------------------------------------------
README				(this file)
Makefile			(for use with make to build)

---------------------------------------------------------------------------------------------------
* fvarsample code (C code functions only - requires GSL www.gnu.org/software/gsl)
fvarsample.c                                    C code for sampling sentiments
fvarsample.h 					C header file

* planning code
forvalue.c					main planning functions
forvalue.h

* x-state structures
defaultAPP.{c,h}				default (stub) functions for a class with no X state
pdAPP.{c,h}					example for the prisoner's dilemma - x state is a single integer

* python wrapper
pywrapper.c                                     functions to allow calling from python3

* auxiliary files
inputparser.c					parser for the input file (see testsim.txt)
inputparser.h					header for the parser
tinycsv.{c,h}					functions for parsing csv files

* test main functions
testbayesact.c 					full simulation with parsed simulation specification file (see README)
testbayesact1.c 				with hardcoded simulation set-up, parser not used
testfvarsample.c				base sampling only
pdbot.py  					Python3 code for PD bot example - this is an example of how to use the python wrapper

---------------------------------------------------------------------------------------------------
* input files
testsim-interact.txt 	An example of how to specify input for a simple mimic-interact simulation
testsim.txt

testsim-pd.txt		An example for the prisoner's dilemma game


---------------------------------------------------------------------------------------------------
* ACT files included with this distribution (see terms of use at http://www.indiana.edu/~socpsy/ACT/legal.htm)
fidentities.dat		        (identity dictionary from Indiana2002-04 database from interact software)
fbehaviours.dat		        (behaviour dictionary from Indiana2002-04 database from interact software)
fmodifiers.dat		      	(modifier/emotions dictionary from Indiana2002-04 database from interact software)
fsettings.dat			(settings dictionary from Indiana2002-04 database from interact software)
tdynamics-male.dat		(equations for dynamics - male - U.S.A. 1978 database)
tdynamics-female.dat		(equations for dynamics - female - U.S.A. 1978 database)
temotions-male.dat		(equations for emotions - male - Indiana2002-2004 database from interact software)
temotions-female.dat		(equations for emotions - female - Indiana2002-2004 database from interact software)

identityEPAandCOV.csv		(identity dictionary from the 2015 Georgia-Duke survey with full covariance matrices)
identityEPAandSD.csv		(identity dictionary from the 2015 Georgia-Duke survey with standard deviations only)
behaviourEPAandCOV.csv		(behaviour dictionary from the 2015 Georgia-Duke survey with full covariance matrices)
behaviourEPAandSD.csv		(behaviour dictionary from the 2015 Georgia-Duke survey with standard deviations only)
modifierEPAandCOV.csv		(modifier dictionary from the 2015 Georgia-Duke survey with full covariance matrices)
modifierEPAandSD.csv		(modifier dictionary from the 2015 Georgia-Duke survey with standard deviations only)

pdidentities.dat		a shortened version of fidentities.dat for use with the PD game
pdidentities.csv		a shortened version of identityEPAandCOV.csv for use with the PD game
pdbehaviours.dat		a shortened version of fbehaviours.dat for use with the PD game
pdbehaviours.csv		a shortened version of behaviourEPAandCOV.csv for use with the PD game
pdemotions.dat			a shortened version of fmodifiers.dat for use with the PD game



