/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.2)                                #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2021                                                                            #
#  Copyright (C) 2021  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- stub version
#************************************************************************************************/


#include "bayesactAPP.h"


// initialize and free

void initializeXXObs(xostate **xobs)
{
  // need to have this to maintain consistency, but it is not used
  *xobs = (xostate *) malloc(sizeof(xostate));
  (*xobs)->dummy = 0;
}
void freeXXObs(xostate **xobs)
{
  free(xobs);
}
// update any additional xstate based on xstate
// and on the actions taken by the two agents
// which are the indices of the denotative actions taken by agent and client
// in the dictionaries specified as BEHAVIOUR in the input file
// generally this is a draw from P(x'|x'_{ba},x'_{bc})
void updateXXSample(gsl_rng *r, xxstate *next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour)
{
  // a stub example
  return;
}


// this is a draw from P(x'_{bc}|x_x, x_bc, x_ba, x_a, x_c) where x_x is in current_xxstate and
void updateSampleXbc(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (o_bc >= 0) {
    (*x_bc)=o_bc;
  } else {
    // irrelevant for getAction - really this plays no role anywhere other than as the expectation of what the client will do,
    // which will be filled in anyways 
    // filter by institution here!
    // something like drawRandomFiltered(the_agent->entity_dicts[BEHAVIOUR],the_agent->institution_dicts[BEHAVIOUR]);
    // this will select a random element from entity_dicts that has a flag set in institution_dicts
    // the flag has a number which in binary is the mask for the institution list.
    //int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_bc) = gsl_rng_uniform_int(r, num_behaviours);

    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_bc) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];

    
  }
}
// this is Pr(x_{ba}'|x,x_{ba})
void updateSampleXba(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (a >= 0) {
    (*x_ba)=a;
  } else {
    // want to draw a sample from the denotative distribution over actions
    //    FILTER BY INSTITUTION HERE!
    //    int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_ba) = gsl_rng_uniform_int(r, num_behaviours);
    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_ba) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];
    
  }
}
// this is Pr(xa',xc'|xa,xc,agent_beh,client_beh)
void updateSampleXac(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int client_behaviour, dyad *the_dyad, agentState *the_agent)
{
  // the identity stays the same by default
  (*x_ap) = x_a;
  (*x_cp) = x_c;

  double AGENT_DELTA=the_agent->deltaa;  //DEFAULT_AGENT_DELTA;
  double CLIENT_DELTA=the_agent->deltac;  //DEFAULT_AGENT_DELTA;
  
  //this value should be a property of the_agent once the ssstate becomes part of the dyad
  // or is accessible through the dyad
  // This is part of an agent's delta value.
  // the other part is the DISRUPT_NOISE
  
  double theid[SSD];
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[AGENT]->theDictionary[x_a].epa,i)+
      gsl_ran_gaussian(r,AGENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_ap, theid, the_agent->entity_dicts[AGENT]);
  
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[CLIENT]->theDictionary[x_c].epa,i)+
      gsl_ran_gaussian(r,CLIENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_cp, theid, the_agent->entity_dicts[CLIENT]);


  return;
}



// this is R(x)
double getXXUtility(xxstate * xx_state)
{
  return 0.0;
}


// weight x' based on o_x'
// generally this returns P(o_x | x')
double weightXXSample(gsl_rng *r, xxstate * xx_state, xostate * xx_observation)
{
  return 1.0;
}
// weight the x_bc sample according to the observation - by default it should be identical
// however, there could be some observation noise that comes into play here. It would also have
// to be present in the weighting of the XX sample though
// weight x_{bc}' based on \omega_{x_{bc}} - this is P(o'_bc | x'_bc)
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc)
{
  if ((*x_bc)==o_bc)
    return 1.0;
  return 0.0;
}
// draw a sample from P(o_x | x) 
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation)
{
  xx_observation = NULL;
}
// draw a sample from P(o_bc | x_bc) 
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc)
{
  // could add noise here
  (*o_bc) = x_bc;
}
//----------------------------------------------------------------------------------------------------
// other helper functions
//----------------------------------------------------------------------------------------------------


// function convert a string to an xx_observation (for user input)
int stringToXXObs(char * str, xostate *xx_observation)
{
  return 1;  // success
}
void xxObsToString(char * retstr,xostate *xx_observation)
{
  strcpy(retstr,"");
}
int compareXXObs(xostate * xx1, xostate *xx2)
{
  return 1;
}

// function to print the xx state and observatoin
void printRawXXstate(xxstate * xx)
{
  return;
}
void printRawXXObs(xostate * xx_observation)
{
  return;
}

// functions to initialize, add, divide and copy XX states
void initializeXXstate(xxstate *xx)
{
  return;
}

void addToXXstate(xxstate *xx_avg, xxstate *xx, double weight)
{
  return;
}
void divideXXstate(xxstate * xx,double v)
{
  return;
}
void copyXXstate(xxstate *dest, xxstate *src)
{
  return;
}

