/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- stub version
#************************************************************************************************/
#ifndef __BAYESACTAPP
#define __BAYESACTAPP

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_rng.h>


// these structs must come before include-ing fvarsample.h
// additional state X
typedef struct xx_state xxstate;
struct xx_state {
	int dummy;
};

// additional state O_x
typedef struct xo_state xostate;
struct xo_state {
	int dummy;
};

#include "fvarsample.h"

// functions - important ones
// initialize and free
void initializeXXObs(xostate **xobs);
void freeXXObs(xostate **xobs);
// update samples
void updateXXSample(gsl_rng *r, xxstate * next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour);
void updateSampleXba(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int, dyad *, agentState *);
void updateSampleXbc(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int, dyad *, agentState *);
void updateSampleXac(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int, dyad *, agentState *);
double weightXXSample(gsl_rng *r, xxstate * xx_state, xostate * xx_observation);
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc);
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc);
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation);
double getXXUtility(xxstate * xx_state);
void initializeXXstate(xxstate *xx);
void copyXXstate(xxstate *dest, xxstate *src);

int compareXXObs(xostate * xx1, xostate *xx2);

// functions - for input and output
int stringToXXObs(char * str, xostate *xx_observation);
void xxObsToString(char * retstr,xostate *xx_observation);
void printRawXXstate(xxstate * xx);
void printRawXXObs(xostate * xx_observation);
void addToXXstate(xxstate *xx_avg, xxstate *xx, double weight);
void divideXXstate(xxstate * xx,double v);
#endif
