/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
# poorly hacked together image writing tools - for writing out images of sample sets
# this needs to be replaced with something better!
#************************************************************************************************/
#include "imagewriter.h"


static int write_png_file(char *filename, unsigned char *buffer, int ncols, int nrows)
{
  FILE *fp = NULL;
  png_structp png_ptr = NULL;
  png_infop info_ptr = NULL;
  
  // Open file for writing (binary mode)
  fp = fopen(filename, "wb");
  if (fp == NULL) {
    fprintf(stderr, "Could not open file %s for writing\n", filename);
    return 1;
  }
  
  // Initialize write structure
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL) {
    fprintf(stderr, "Could not allocate write struct\n");
    return 1;
  }

  // Initialize info structure
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == NULL) {
    fprintf(stderr, "Could not allocate info struct\n");
    return 1;
  }

  // Setup Exception handling
  if (setjmp(png_jmpbuf(png_ptr))) {
    fprintf(stderr, "Error during png creation\n");
    return 1;
  }
  
  png_init_io(png_ptr, fp);
  
  // Write header (8 bit colour depth)
  png_set_IHDR(png_ptr, info_ptr, ncols, nrows,
	       8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	       PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  
  png_write_info(png_ptr, info_ptr);

  // Write image data
  for (int y=0 ; y<nrows ; y++) 
    png_write_row(png_ptr, buffer+(y*ncols*3));
  
  // End write
  png_write_end(png_ptr, NULL);
  
  if (fp != NULL) fclose(fp);
  if (info_ptr != NULL) png_free_data(png_ptr, info_ptr, PNG_FREE_ALL, -1);
  if (png_ptr != NULL) png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
  
  return 0;
}

void save_ppm(char *path, unsigned char *data, int ncols, int nrows)
{
  FILE* stream  = fopen(path,"w");
  if ( !stream )
    return;

  fprintf(stream, "P6\n%d %d\n255\n", ncols, nrows );
  
  for ( int r = 0; r < nrows; r++ )
    for ( int c = 0; c < ncols; c++ ) {
      fwrite(data, 1, 1, stream);
      fwrite(data+1, 1, 1, stream);
      fwrite(data+2, 1, 1, stream);
      data += 3;
    }
  fclose( stream );
}
static void draw_line(unsigned char *image, int ncols, int nrows, int x, int y, int xt, int yt, int red, int green, int blue)
{
  double llen = sqrt((xt-x)*(xt-x)+(yt-y)*(yt-y));
  int ilen = (int) ceil(llen);
  double xstep = (1.0*(xt-x))/(1.0*ilen);
  double ystep = (1.0*(yt-y))/(1.0*ilen);
  double dr=y;
  double dc=x;
  int r,c, index;
  for (int j=0; j<ilen; j++) {
    dr += ystep;
    dc += xstep;
    r = (int) ceil(dr);
    c = (int) ceil(dc);
    index = MAX(0,MIN(nrows-1,(r)))*ncols+MAX(0,MIN(ncols-1,c));
    image[index*3]=red;
    image[index*3+1]=green;
    image[index*3+2]=blue;
  }
}
static void create_sample_image(unsigned char * image, int ncols, int nrows, sample *samples, int numsamples, pair_type type)
{
  // type is a binary xxx = epa 
  // 110 = 6 is e vs p  (EP)
  // 101 = 5 is e vs a  (EA)
  // 011 = 3 is p vs a  (PA)
  int index,indexc,indexa;
  unsigned char full=(unsigned char) 255;
  double e,p,a,ec,pc,ac;
  double et,pt,at,ect,pct,act;
  int r,c,i;
  int x,y,xc,yc;
  int xt,yt,xct,yct;
  int sz=1;
  double maxv=4.3;
  double maxv2=maxv*2;
  int max_draw_samples=250;
  int skip=1;
  if (numsamples > max_draw_samples) {
    skip = numsamples/max_draw_samples;
  }
  // draw a line from fundamental to transient?
  // set this to 0 for no line, and to 255 for a VERY faint line
  int doline=128;
  //fprintf(stderr,"skip is %d\n",skip);
  for (i=0; i<numsamples; i++) {
    if (samples[i].weight < 1.0) {
      fprintf(stderr,"Unable to write image - weights have to be 1.0\n");
      return;
    }
  }
  
  double savesz[nrows*ncols];
  int maxx, maxy, minx, miny;
  double dist_from_cent;
  int do_radial=1;  // flag - if 0 will be "blocky"
  // this can be bigger if the samples are more identical
  double sz_fudge_factor=0.0001;
  for ( r = 0; r < nrows; r++ )
    for ( c = 0; c < ncols; c++ ) {
      index = (r*ncols+c)*3;
      image[index]=full;
      image[index+1]=full;
      image[index+2]=full;
      savesz[(r*ncols+c)]=sz*2;
    }



  // connect fundamentals to transients with a line
  if (doline>0) {
    for (i=0; i<numsamples; i+=skip) {
      // agent samples
      e=MAX(0,MIN(1,(samples[i].f[0]+maxv)/(maxv2)));
      p=MAX(0,MIN(1,(samples[i].f[1]+maxv)/(maxv2)));
      a=MAX(0,MIN(1,(samples[i].f[2]+maxv)/(maxv2)));
      
      // client
      ec=MAX(0,MIN(1,(samples[i].f[6]+maxv)/(maxv2)));
      pc=MAX(0,MIN(1,(samples[i].f[7]+maxv)/(maxv2)));
      ac=MAX(0,MIN(1,(samples[i].f[8]+maxv)/(maxv2)));
      
      if (type==PA)  {
	x = (int) ((ncols-1)*p);      y = (int) (nrows-(nrows-1)*a);
	xc = (int) ((ncols-1)*pc);      yc = (int) (nrows-(nrows-1)*ac);
      }
      if (type==EA)  {
	x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*a);
	xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*ac);
      }
      if (type==EP)  {
	x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*p);
	xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*pc);
      }
      
      // agent samples
      et=MAX(0,MIN(1,(samples[i].tau[0]+maxv)/(maxv2)));
      pt=MAX(0,MIN(1,(samples[i].tau[1]+maxv)/(maxv2)));
      at=MAX(0,MIN(1,(samples[i].tau[2]+maxv)/(maxv2)));
      
      // client
      ect=MAX(0,MIN(1,(samples[i].tau[6]+maxv)/(maxv2)));
      pct=MAX(0,MIN(1,(samples[i].tau[7]+maxv)/(maxv2)));
      act=MAX(0,MIN(1,(samples[i].tau[8]+maxv)/(maxv2)));
      
      if (type==PA)  {
	xt = (int) ((ncols-1)*pt);      yt = (int) (nrows-(nrows-1)*at);
	xct = (int) ((ncols-1)*pct);      yct = (int) (nrows-(nrows-1)*act);
      }
      if (type==EA)  {
	xt = (int) ((ncols-1)*et);      yt = (int) (nrows-(nrows-1)*at);
	xct = (int) ((ncols-1)*ect);      yct = (int) (nrows-(nrows-1)*act);
      }
      if (type==EP)  {
	xt = (int) ((ncols-1)*et);      yt = (int) (nrows-(nrows-1)*pt);
	xct = (int) ((ncols-1)*ect);      yct = (int) (nrows-(nrows-1)*pct);
      }
      
      // now we want to draw a line, in very light red, from x,y -- xt,yt
      // and another one, in blue, from xc,yc -- xct,yct
      draw_line(image,ncols,nrows,x,y,xt,yt,255,doline,doline);
      draw_line(image,ncols,nrows,xc,yc,xct,yct,doline,doline,255);
    }
  }





  
  for (i=0; i<numsamples; i+=skip) {
    // agent samples
    e=MAX(0,MIN(1,(samples[i].f[0]+maxv)/(maxv2)));
    p=MAX(0,MIN(1,(samples[i].f[1]+maxv)/(maxv2)));
    a=MAX(0,MIN(1,(samples[i].f[2]+maxv)/(maxv2)));

    // client
    ec=MAX(0,MIN(1,(samples[i].f[6]+maxv)/(maxv2)));
    pc=MAX(0,MIN(1,(samples[i].f[7]+maxv)/(maxv2)));
    ac=MAX(0,MIN(1,(samples[i].f[8]+maxv)/(maxv2)));

    if (type==PA)  {
      x = (int) ((ncols-1)*p);      y = (int) (nrows-(nrows-1)*a);
      xc = (int) ((ncols-1)*pc);      yc = (int) (nrows-(nrows-1)*ac);
    }
    if (type==EA)  {
      x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*a);
      xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*ac);
    }
    if (type==EP)  {
      x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*p);
      xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*pc);
    }
    // get index in image where sample will go
    index = MAX(0,MIN(nrows-1,y))*ncols+MAX(0,MIN(ncols-1,x));
    // grow one bigger -- so it will get sz pixels bigger each time a sample is added.d
    // this can rapidly block out an image, so a factor is added
    savesz[index] = MIN(MAX_DOT_SIZE,savesz[index]+log10(savesz[index]+sz_fudge_factor*sz));
    // agent is red - bigger means more samples at that point
    miny=minx=ceil(-savesz[index]);
    maxy=maxx=floor(savesz[index]);
    for (r=miny; r<=maxy; r++) 
      for (c=minx; c<=maxx; c++) {
	dist_from_cent=sqrt(1.0*((c*c)+(r*r)));
	if (!do_radial || dist_from_cent < savesz[index]) {
	  indexa = MAX(0,MIN(nrows-1,(y+r)))*ncols+MAX(0,MIN(ncols-1,x+c));
	  image[indexa*3]=255;  // this makes it darker as it gets bigger : MAX(MIN_COL,image[indexa*3]-10);
	  image[indexa*3+1]=0;
	  image[indexa*3+2]=0;
	}
      }
    // also grow bigger

    // client is blue
    // get index in image where sample will go
    index = MAX(0,MIN(nrows-1,yc))*ncols+MAX(0,MIN(ncols-1,xc));
    // grow one bigger -- so it will get sz pixels bigger each time a sample is added.d
    // this can rapidly block out an image, so a factor is added
    // note there is only one savez[index] for each pixel ... !
    savesz[index] = MIN(MAX_DOT_SIZE,savesz[index]+log10(savesz[index]+sz_fudge_factor*sz));
    // agent is red - darker and bigger means more samples at that point
    miny=minx=ceil(-savesz[index]);
    maxy=maxx=floor(savesz[index]);
    for (r=miny; r<=maxy; r++) 
      for (c=minx; c<=maxx; c++) {
	dist_from_cent=sqrt(1.0*((c*c)+(r*r)));
	if (!do_radial || dist_from_cent < savesz[index]) {
	  indexc = MAX(0,MIN(nrows-1,(yc+r)))*ncols+MAX(0,MIN(ncols-1,xc+c));
	  //if (image[indexc*3]==255)
	  image[indexc*3]=0;
	  image[indexc*3+1]=0;
	  image[indexc*3+2]=255; // MAX(MIN_COL,image[indexc*3+2]-10);
	}
      }
  }
  // transients - as crosses
  for (i=0; i<numsamples; i+=skip) {
    // agent samples
    e=MAX(0,MIN(1,(samples[i].tau[0]+maxv)/(maxv2)));
    p=MAX(0,MIN(1,(samples[i].tau[1]+maxv)/(maxv2)));
    a=MAX(0,MIN(1,(samples[i].tau[2]+maxv)/(maxv2)));

    // client
    ec=MAX(0,MIN(1,(samples[i].tau[6]+maxv)/(maxv2)));
    pc=MAX(0,MIN(1,(samples[i].tau[7]+maxv)/(maxv2)));
    ac=MAX(0,MIN(1,(samples[i].tau[8]+maxv)/(maxv2)));

    if (type==PA)  {
      x = (int) ((ncols-1)*p);      y = (int) (nrows-(nrows-1)*a);
      xc = (int) ((ncols-1)*pc);      yc = (int) (nrows-(nrows-1)*ac);
    }
    if (type==EA)  {
      x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*a);
      xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*ac);
    }
    if (type==EP)  {
      x = (int) ((ncols-1)*e);      y = (int) (nrows-(nrows-1)*p);
      xc = (int) ((ncols-1)*ec);      yc = (int) (nrows-(nrows-1)*pc);
    }
    // get index in image where sample will go
    //index = MAX(0,MIN(nrows-1,y))*ncols+MAX(0,MIN(ncols-1,x));
    // grow one bigger -- so it will get sz pixels bigger each time a sample is added.d
    // this can rapidly block out an image, so a factor is added
    //savesz[index] = savesz[index]+MIN(MAX_DOT_SIZE,log10(savesz[index]+sz_fudge_factor*sz));
    // agent is red - darker and bigger means more samples at that point
    //miny=minx=ceil(-savesz[index]);
    //maxy=maxx=floor(savesz[index]);
    maxy=MIN(MAX_DOT_SIZE,savesz[index]+log10(savesz[index]+sz_fudge_factor*sz));
    miny=-maxy;
    //miny=minx=ceil(-sz);
    //maxy=maxx=floor(sz);
    // agent is red - darker means more samples at that point
    for (r=miny; r<=maxy; r++) 
      for (c=minx; c<=maxx; c++) {
	if (r==0 || c==0) { // to make a cross
	  //dist_from_cent=sqrt(1.0*((c*c)+(r*r)));
	
	  //if (!do_radial || dist_from_cent < savesz[index]) {
	  indexa = MAX(0,MIN(nrows-1,(y+r)))*ncols+MAX(0,MIN(ncols-1,x+c));
	  image[indexa*3]=255; //MAX(MIN_COL,image[indexa*3]-10);
	  image[indexa*3+1]=0;
	  image[indexa*3+2]=0;
	  //}
	  }
      }
    // client is blue
    for (r=miny; r<=maxy; r++) 
      for (c=minx; c<=maxx; c++) {
	if (r==0 || c==0) { // to make a cross
	  //dist_from_cent=sqrt(1.0*((c*c)+(r*r)));
	  //if (!do_radial || dist_from_cent < savesz[index]) {
	    indexc = MAX(0,MIN(nrows-1,(yc+r)))*ncols+MAX(0,MIN(ncols-1,xc+c));
	    //if (image[indexc*3]==255)
	    image[indexc*3]=0;
	    image[indexc*3+1]=0;
	    image[indexc*3+2]=255; //MAX(MIN_COL,image[indexc*3+2]-10);
	    //}
	}
      }
  }
  // draw axes
  draw_line(image,ncols,nrows,0,(int) (0.5*(nrows-1)),ncols-1,(int) (0.5*(nrows-1)),0,0,0);
  draw_line(image,ncols,nrows,(int) (0.5*(ncols-1)),0,(int) (0.5*(ncols-1)),nrows-1,0,0,0);
}

// last argument here is type of image: EP, EA or PA
int write_sample_image(char *basefname,int dyad, int iteration, int ncols, int nrows, sample *samples, int numsamples, pair_type type)
{
  unsigned char *image=(unsigned char*) malloc(nrows*ncols*3*sizeof(unsigned char));
  char *path= malloc(256*sizeof(char));
  sprintf(path,"%s%d-%d.png",basefname,dyad,iteration);

  create_sample_image(image,ncols,nrows,samples,numsamples,type);
  //save_ppm(path,image,ncols,nrows);
  int retval=0;
  if (write_png_file(path,image,ncols,nrows)) 
    retval=1;
  free(path);
  free(image);
  return retval;
}

