/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.2)                                #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2021                                                                            #
#  Copyright (C) 2017-2021  Jesse Hoey                                                           #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
# poorly hacked together parser -- needs to be replaced with something better
# see example of what this reads in testsim.txt
#************************************************************************************************/
#include "inputparser.h"

// returns a pointer to the dictionary that comes from filename fname in the list dictionaries
// or NULL if it is not in the list
static dictionary * get_dictionary(dictionary * dictionaries, int num_dictionaries, char *fname)
{
  int k=0;
  while (k<num_dictionaries) {
    if (strcmp(dictionaries[k].fname,fname)==0) {
      return dictionaries+k;
    }
    k++;
  }
  return NULL;
}
// return the index of agent name from the list agents
static int find_agent(agentState *agents, int num_agents, char *name)
{
  for (int k=0; k<num_agents; k++)
    if (strcmp(agents[k].name,name)==0)
      return k;
  return -1;
}

// given a list of strings columnX with num_token elements, read alternating
// identity labels and probabilities and assign each to an element of agent_ids
// labels are assigned indices in the id_dictionary of agent
// assumes the first element of columnX contains the agent name and will be ignored
static int extract_ids(int num_tokens, identity **agent_ids, char ** columnX)
{
  int i,k;
  int num_ids = (num_tokens-1)/2;
  if (num_ids==0) {
    // unknown identity
    num_ids=-1;
    *agent_ids=NULL;
  } else {
    *agent_ids = (identity *) malloc(num_ids*sizeof(identity));
    // self identity
    i=0;
    double total_prob=0.0;
    for (k=1; k<num_tokens; k+=2) {
      strcpy((*agent_ids)[i].label,columnX[k]);
      (*agent_ids)[i].prob=atof(columnX[k+1]);
#ifdef VERBOSE
      printf("identity %s probability %4.3g\n",columnX[k],(*agent_ids)[i].prob);
#endif
      total_prob += (*agent_ids)[i].prob;
      i++;
    }
    if (total_prob-1.0 > 0.0) {
      printf("probabilities don't sum to 1\n");
      exit(0);
    }
  }
  return num_ids;
}

// read events from a file
// from events file
// tt==0 is synonymous with the AGENT of this dyad,
// which is the AGENT of the first dyad from the events file
// if I was doing the mirror dyad (events[iteration].dyad_index[1]), then
// this is the CLIENT of the first dyad from the events file
// for now it works so long as all the events in the events file
// start with the name of the agent of the dyad we want updated
// so if you have
// interaction: hank : tom
// and
// intearction: tom :hank
// in the simulation file, then in the events file
// hank : * : : tom :  : :
// means hank in dyad 1 (this agent hank from  from hank's perspective in the hank-tom interaction dyad) is doing optimal,
// while tom in dyad 1 (this is agent tom from hank's perspective in the hank-tom interaction dyad) is doing nothing
// whereas
// tom :  : : hank :* : : 
// means hank in dyad 2 (this agent hank from  from tom's perspective in the tom-hank interaction dyad) is doing optimal (from TOM's perspective),
// while tom in dyad 2 (this is agent tom from tom's perspective in the tom-hank interaction dyad) is doing nothing
//
// the keyword "interaction" can now be followed by a line
// institution:
// with a :-separated list of institutions in
// overt, surmised, lay, business, law, politics, academe, medicine, religion, family, sexual, monadic, group, corporal
//
int readEvents(char *events_fname, dyad *dyads, int num_dyads, agentState *agents, int num_agents, event **events)
{
  int max_iterations=0;
  int k;
  int event_num=0;
  int dyad_index;
  int agent_index, agent_behaviour_index, agent_emotion_index;
  int client_index, client_behaviour_index, client_emotion_index;
  int disrupt_action[2];
  static int max_line_length=2048;
  static int max_num_tokens=100;  // must be bigger than expected
  FILE * fp = fopen(events_fname,"r");
  char buffer[max_line_length];
  char agent_x_observation[max_line_length];
  char client_x_observation[max_line_length];
  if (!fp) 
    return 0;


  // find out number of iterations
  int num_tokens, line_length;
  while(fgets(buffer, max_line_length, (FILE*) fp)) {
    if (!(buffer[0]=='/' && buffer[1]=='/')) {  // skip comments
      line_length=strlen(buffer);
      if (line_length>1)
	max_iterations++;
    }
  }
#ifdef VERBOSE
  printf("max iterations: %d\n",max_iterations);
#endif
  *events = (event *) malloc(max_iterations*sizeof(event));
  rewind(fp);
  while(fgets(buffer, max_line_length, (FILE*) fp)) {
    if (!(buffer[0]=='/' && buffer[1]=='/') && !(buffer[0]=='\n')) {  // skip comments and empty lines
      line_length=strlen(buffer);
      // remove trailing newline
      buffer[strcspn(buffer, "\n")] = 0;
      char **columnX = NULL;
      num_tokens = getcols(buffer, ":", &columnX, max_num_tokens);
      for (k=0; k<num_tokens; k++) 
	trim(columnX[k]);
      //====================================================
      // start processing the input elements
      //====================================================
      // each line should have 6 elements maximum
      // 0         1                 2               3        4                 5            6
      // agent : agent_behaviour : agent_emotion : client [client_behaviour client_emotion : x_observation ]
      // if client_behaviour is missing, it is assume to be "none" (client does not act)
      // if agent_behaviour is missing, client behaviour must be there
      // if both behaviours are missing, it is an error
      // if a behaviour or emotion is * it will be the default (most likely bayesact) action when simulating 
      // if a behaviour ! it will be the interact action when simulating
      // if any behaviour slot (even empty) has "+" at the end,
      // it has added random noise of amount DISRUPT_NOISE for the observer of the behaviour only
      // if agent_emotion or client_emotion is missing, they are assumed to be "no emotion"
      agent_index = find_agent(agents,num_agents,columnX[0]);
      //printf("agent: %s index %d\n",columnX[0],agent_index);

      client_index = find_agent(agents,num_agents,columnX[3]);
      //printf("client: %s %d\n",columnX[3],client_index);

      // find which dyad this is
      dyad_index=-1;
      for (k=0; k<num_dyads; k++) {
	if (dyads[k].agent_index==agent_index && dyads[k].client_index==client_index)
	  dyad_index=k;
	// inverse - this is set up when dyads are created now using mirror_dyad_index of dyadType
	//if (dyads[k].agent_index==client_index && dyads[k].client_index==agent_index)
	//dyad_index[1]=k;
      }
      // exit if dyad not found
      if (dyad_index<0) {
	printf("Dyad not found in events file %s\n",columnX[1]);
	exit(0);
      }
      // this means we will add randomness to the behaviour
      disrupt_action[ID_AGENT]=0;
      if (columnX[1][strlen(columnX[1])-1]=='+') {
	disrupt_action[ID_AGENT]=1;
	columnX[1][strlen(columnX[1])-1]='\0';
      }
      if (strlen(columnX[1])==0) {      // did not act
	agent_behaviour_index=NO_ACT;
      } else if (columnX[1][0]=='*') {  // use BayesACT optimal
	agent_behaviour_index=BAYES_ACT;
      } else if (columnX[1][0]=='!') {    // use interact optimal
	agent_behaviour_index=INTER_ACT;       
#ifndef DO_INTERACT	  
	fprintf(stderr,"DO_INTERACT MUST be define-d in fvarsample.h if you want to use INTERACT actions!\n");
	exit(0);
#endif
      } else {
	// look it up
	agent_behaviour_index=lookUpLabel(columnX[1],agents[agent_index].entity_dicts[BEHAVIOUR]);
	if (agent_behaviour_index < 0) {
	  printf("behaviour not found in events file %s\n",columnX[1]);
	  exit(0);
	}
      }
	
      //printf("agent behaviour %s %d\n",columnX[1],agent_behaviour_index);

      if (strlen(columnX[2])==0) {
	agent_emotion_index=-1;
      } else if (columnX[2][0]=='*') {
	agent_emotion_index=-2;
      } else {
	// look it up
	agent_emotion_index=lookUpLabel(columnX[2],agents[agent_index].entity_dicts[EMOTION]);
	if (agent_emotion_index < 0) {
	  printf("emotion not found in events file %s\n",columnX[2]);
	  exit(0);
	}
      }
      //printf("agent emotion %s %d\n",columnX[2],agent_emotion_index);

      // client behaviours and emotions  and x observation ----------------
      if (num_tokens > 4) {
	// this means we will add randomness to the behaviour
	disrupt_action[ID_CLIENT]=0;
	if (columnX[4][strlen(columnX[4])-1]=='+') {
	  disrupt_action[ID_CLIENT]=1;
	  columnX[4][strlen(columnX[4])-1]='\0';
	}

	// client behaviour
	if (strlen(columnX[4])==0) {
	  client_behaviour_index=NO_ACT;
	} else if (columnX[4][0]=='*') {    // use bayesact optimal
	  client_behaviour_index=BAYES_ACT;
	} else if (columnX[4][0]=='!') {    // use interact optimal
	  client_behaviour_index=INTER_ACT;

#ifndef DO_INTERACT	  
	  fprintf(stderr,"DO_INTERACT MUST be define-d in fvarsample.h if you want to use INTERACT actions!\n");
	  exit(0);
#endif

	} else {
	  // look it up
	  client_behaviour_index=lookUpLabel(columnX[4],agents[client_index].entity_dicts[BEHAVIOUR]);
	  if (client_behaviour_index < 0) {
	    printf("behaviour not found in events file %s\n",columnX[4]);
	    exit(0);
	  }
	}
	//	printf("client behaviour %s %d\n",columnX[4],client_behaviour_index);
      } else {
	client_behaviour_index=-1;
      }
      if (num_tokens > 5) {
	if (strlen(columnX[5])==0) {
	  client_emotion_index=-1;
	} else if (columnX[5][0]=='*') {
	  client_emotion_index=-2;
	} else {
	  // look it up
	  client_emotion_index=lookUpLabel(columnX[5],agents[client_index].entity_dicts[EMOTION]);
	  if (client_emotion_index < 0) {
	    printf("emotion not found in events file %s\n",columnX[5]);
	    exit(0);
	  }
	}
	//printf("client emotion %s %d\n",columnX[5],client_emotion_index);
      } else {
	client_emotion_index=-1;
      }
      if (num_tokens > 6) {
	// this is "x_observation" which is interpreted by the application being used
	// so it remains a string
	strcpy(agent_x_observation,columnX[6]);
      }
      if (num_tokens > 7) {
	// this is "x_observation" which is interpreted by the application being used
	// so it remains a string
	strcpy(client_x_observation,columnX[7]);
      }
	
      //printf("\n");
      free(columnX);
      // set up the event:
      (*events)[event_num].dyad_index=dyad_index;
      (*events)[event_num].agent_index=agent_index;
      (*events)[event_num].client_index=client_index;
      (*events)[event_num].behaviour_index[ID_AGENT]=agent_behaviour_index;
      (*events)[event_num].behaviour_index[ID_CLIENT]=client_behaviour_index;
      (*events)[event_num].emotion_index[ID_AGENT]=agent_emotion_index;
      (*events)[event_num].emotion_index[ID_CLIENT]=client_emotion_index;
      (*events)[event_num].disrupt_action[ID_AGENT]=disrupt_action[ID_AGENT];
      (*events)[event_num].disrupt_action[ID_CLIENT]=disrupt_action[ID_CLIENT];
      if (num_tokens>6) 
	strcpy((*events)[event_num].agent_x_obs,agent_x_observation);
      if (num_tokens>7)
	strcpy((*events)[event_num].client_x_obs,client_x_observation);
      event_num++;
    }
  }
  fclose(fp);
  return max_iterations;
}
// do all the setup based on simulation file fname
//
// input: sstate is the samplerState that will contain all the samples and parameters for each interaction
// input: fname is the input file (see example and notes in testsim.txt)
// output: agents are the agentState agents involved in the interaction - includes pointers to dictionaries for all entity types
// output: num_agents is the number of agents involved
// output: dyads is a list of dyad structures, each corresponding to an interaction (a single agent-client dyad)
//         and giving an index into sstate->samples  and sstate->sentiParams, and an index for agent and client in that interaction
// output: num_dyads is the number of dyads 
// output: query_user is 0 (BAYESACT) if bayesact should be used for action selection, 1 (USER) if user should be queried in "interactive" mode
//         if 2 (EVENTS), then events are read from an events file
// output: max_iterations is the number of iterations to do
// output: dictionaries points to a list of num_dictionaries dictionary structures "dictionaries" these agents use
// output: num_dictionaries is the number of dictionaries read in
// returns 0 on success, otherwise returns -1
int setUpAgentsDyads(char *fname, agentState * agents, int * num_agents,
		     dyad *dyads, int *num_dyads, query_type * query_user, int * max_iterations, double gamma,
		     dictionary * dictionaries, int * num_dictionaries, event **events)
{
  char * events_fname = (char *) malloc(MAX_LABEL_LENGTH*sizeof(char));
  // this reads the simulation specification from fname, and initializes all agents, dictionaries and sentiment Parameters and samples
  if (readSimulationInput(fname, agents, num_agents, dyads, num_dyads, query_user, max_iterations, &events_fname)<0) {
    printf("error in simulation input file %s\n",fname);
    return -1;
  }
#ifdef VERBOSE
  printf("number of agents read in: %d\n",*num_agents);
#endif

  //read in dictionaries from the names found for each agent
  //gamma is read is as default here for each dictionary, then overriden below with gamma input to this function
  assign_dictionaries(agents,*num_agents, dictionaries, num_dictionaries);

  // lock in value of gamma, if specified, as the variance on each entry in each dictionary, 
  // otherwise these variances will be DYTPE_MEAN_VARIANCE as usual
  // if the dictionary type is not MEAN,
  // then gamma has no effect and is determined on a concept-by-concept
  // basis in the dictionary data. 
  if (gamma >= 0.0) {
    for (int j=0; j < (*num_dictionaries); j++) {
      if (dictionaries[j].dtype==MEAN) {
	dictionaries[j].gamma=gamma; 
	for (int k=0; k < dictionaries[j].num_items; k++) 
	  dictionaries[j].theDictionary[k].var=gamma;
      }
    }
  }


  // ================================================================================================
  // initialize the sampler states for each dyad
  // the actual samples for all agents/dyads live in sstate
  // these is one set of samples for each dyad,
  // which represents a particular agent's interaction with one other agent
  // ================================================================================================
  for (int j=0; j<(*num_dyads); j++) {
#ifdef VERBOSE
    printf("**************************** starting BayesACT for dyad %d\n",j);
#endif
    // first pass on 2.3.8 rehash - we initialize all dyads to have the same exact parameters
    // really I need to include alpha etc int eh dyads and have them assigned in setupAgentsDyads above
    //samplerState * initializeBayesACT(int numsamples, double alphaa, double alphab, double alphac,double betaa, double betac, int do_resets,int default_action, int use_emotions, int emotion_planning);
    // hacked in for now - the values for all these things should be assined to the dyad in each, so this function should look like
    initializeBayesACT(dyads+j,agents+dyads[j].agent_index);
    //dyads[j].dsstate = initializeBayesACT(1000,0.01,0.01,0.01,0.0001,0.0001,1,0,0,0); 
  }
  // a BayesACT agent is a set of dyads, a set of agents, and a set of events,
  // and each dyad has a "samplerState" which includes a set of smaples, which are the state
  // therefore, it is more like a "group" than an agent. An actual BayesACT agent would be comprised
  // of a number of these "group" agents, and might contain two sets of group agents, one "in" group and
  // one "out" group. One I define the superclass of BayesACT "Full Actors"  (see schematic diagram)
  // then it will contain numerous of these groups and would be closer to the concept of a full BayesACT agent.


#ifdef VERBOSE
  printf("========================================================================== setting up interactions\n");
#endif
  // set up the interactions for each dyad
  for (int j=0; j<(*num_dyads); j++) {
#ifdef VERBOSE
    printf("dyad %d\n",j);
#endif
    dyads[j].dyad_index=j;  // this is probably irrelevant now
    if (setUpDyadicInteraction(dyads+j, agents+dyads[j].agent_index)<0) {
      return -1;  // error
    }
  }

#ifdef VERBOSE
  printf("========================================================================== setting up mirror dyads\n");
#endif
  // set up the mirror dyads
  for (int j=0; j<(*num_dyads); j++) 
    dyads[j].mirror_dyad_index=-1;
  for (int j=0; j<(*num_dyads); j++) {
    for (int k=j+1; k<(*num_dyads); k++) {
      if (dyads[j].agent_index==dyads[k].client_index && dyads[j].client_index==dyads[k].agent_index) {
	dyads[j].mirror_dyad_index = dyads[k].dyad_index;
	dyads[k].mirror_dyad_index = dyads[j].dyad_index;
      }
    }
  }

  if (strlen(events_fname)>0) {
#ifdef VERBOSE
    printf("events filename: %s\n",events_fname);
#endif
    *max_iterations=readEvents(events_fname,dyads,*num_dyads,agents,*num_agents,events);
    if (*max_iterations <= 0) {
      printf("error in events file\n");
      return -1;
    }
  } else {
    if (*query_user==EVENTS) 
      // defaults to bayesact
      *query_user=BAYESACT;
  }
  free(events_fname);
  // finally we possibly set up institutional filters if they exist for this dictionary/dyad combination
  for (int i=0; i<*num_dictionaries; i++) {
    dictionaries[i].num_ic_items=NULL;
    dictionaries[i].list_ic_converter=NULL;
    if (dictionaries[i].dict_what != EMOTION) 
      buildInstitutionFilterList(dictionaries+i, dyads, *num_dyads);
  } 
  return 0;
}
void printInstitutionFilterList(dictionary *the_dict, dyad *the_dyads, int num_dyads)
{
  if (the_dict->num_ic_items ==NULL)
    return;
  for (int k=0; k<num_dyads; k++) 
    printInstitutionFilterListOneDyad(the_dict,k);
}
void printInstitutionFilterListOneDyad(dictionary *the_dict, int k)
{
  int num_ids;
  if (the_dict->num_ic_items ==NULL)
    return;
  num_ids=the_dict->num_ic_items[k];
  printf("dyad %d, num_ids %d : ",k,num_ids);
  for (int j=0; j<num_ids; j++) 
    printf("%d ",the_dict->list_ic_converter[k][j]);
  printf("\n");
}
  // filter by institution here! 
  // filter by institution here!
//    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
//    int tmp_id=gsl_rng_uniform_int(r,num_ic_identities);
//    (*x_bc) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];
// each dictionary has, for each dyad,  a converter list_ic_converter[dyad_index][i]=j where i=position in the filtered list, and j=position in the original list
// where the filtererd list is the dictionary filtered by the institutional filter specified for the dyad dyad_index
void buildInstitutionFilterList(dictionary *the_dict, dyad *the_dyads, int num_dyads)
{
  
  the_dict->list_ic_converter = (int **) malloc(num_dyads*sizeof(int *));
  the_dict->num_ic_items = (int *) malloc(num_dyads*sizeof(int));

  int num_identities=the_dict->num_items;
  int num_ids=0;
  int l;
  dyad *the_dyad;
  for (int k=0; k<num_dyads; k++) {
    the_dyad=the_dyads+k;
    num_ids=0;
    // first, count the number of items that pass the test
    for (int j=0; j<num_identities; j++) {
      int i=0;
      while ( i<NUM_INSTITUTIONS && ( (the_dict->theDictionary[j].the_institutions[i]+the_dyad->institution_list[i]) < 2 ) )
	i++;
      if (i<NUM_INSTITUTIONS) 
	num_ids++;
    }
    // at this point, num_ids could be zero. If it is, the somatic Transform will simply select from all possible IDS
    // but this could be modified here  by checking if num_ids==0, and if so, setting it to be some specific subset, like what Dave Heise suggestsed
    // Hero, Villain or Fool 
    // Dave Heise email in response to my question about when institutional filters are applied (and what to do if they overly constrain action and identity)
    //
    // First, if reidentification is a problem for agent, see if it works better for client; and vice versa. Second, if there is no strictly institutional identity that is a reasonable fit, go to a pan-institutional identity like hero, villain, or fool. (A classic article on this is Klapp "Heroes, Villains and Fools, as Agents of Social Control," American Sociological Review, Vol. 19, No. 1. [Feb., 1954], pp. 56-62.) Some pan-institutional identity probably is a reasonable fit even if it's not in one's dictionary. This option is interesting because it explains how we expand our notions of self and others. Third, only as a last resort adopt an identity from an irrelevant institution, recognizing that this requires a full redefinition of the situation including reassignment of identities to others and a redefinition of the setting.
    // Lynn Smith-Lovin reply to Heise,
    // But, if you have to choose between a weird institutional identity that is close to the predicted profile (a great-grandmother in a prison) and an appropriate one that is further away (e.g., if there are no non-institutional identities close by)…..I’d accept the distance.  My reading of the results of the new studies is that people are really quite sensitive to the institutional stuff, and institutionally non-sensical identity labelings are just not viewed as plausible.
    //
    // all in all, they seem to indicate that you do what you can to find an institutionally correct identity,
    // sacrificing some optimality in doing so, and falling back on some generic identity if the you really can't find one.
    // Right now, you get a random identity in this case, but remember this is per sample,
    // so overall it is more likely to find the correct identity (or something close)
    the_dict->num_ic_items[k]=num_ids;
    
    the_dict->list_ic_converter[k]= (int *) malloc(num_ids*sizeof(int));
    
    // second fill in the list_ic_converter for this dyad
    l=0;
    for (int j=0; j<num_identities; j++) {
      int i=0;
      while ( i<NUM_INSTITUTIONS && ( (the_dict->theDictionary[j].the_institutions[i]+the_dyad->institution_list[i]) < 2 ) )
	i++;
      if (i<NUM_INSTITUTIONS) {
	if (l>num_ids) {
	  fprintf(stderr,"something wrong in buildInstitutionFilterList! l %d > num_ids %d\n",l,num_ids);
	  exit(0);
	}
	the_dict->list_ic_converter[k][l++]=j;
      }
    }
  }  
}

void freeInstitutionFilterList(dictionary *the_dict, int num_dyads) {
  if (the_dict->num_ic_items == NULL)
    return;
  for (int k=0; k<num_dyads; k++)
    free(the_dict->list_ic_converter[k]);
  free(the_dict->list_ic_converter);
  free(the_dict->num_ic_items);
  the_dict->num_ic_items=NULL;
  the_dict->list_ic_converter=NULL;
  return;
}
       
// reads in a sentiment dictionary from file filename into the_dictionary
int readSentimentDictionary(dict_type dtype, char * filename, dictionary * the_dictionary) {
  char **rows;
  int k;
  the_dictionary->dtype = dtype;
  strcpy(the_dictionary->fname,filename);

  // the default is DTYPE_MEAN_VARIANCE
  if (the_dictionary->dtype == MEAN)
    the_dictionary->gamma=DTYPE_MEAN_VARIANCE;

  // first find out how many there are and what the maximum line length is etc.
#ifdef VERBOSE
  printf("reading from %s\n",the_dictionary->fname);
#endif
  
  static int max_num_tokens=MAX_NUM_TOKENS;   
  static int max_line_length=MAX_LINE_LENGTH;

  FILE * fp = fopen(the_dictionary->fname,"r");
#if UNITY_WIN
  char *buffer = (char *)_alloca(sizeof(char) * max_line_length);
#else
    char buffer[max_line_length];
#endif
  if (!fp)
    return 0;
  rows = NULL; //getcols() requires it to be NULL. (Avoid dangling pointers, leaks e.t.c).

  k=0;
  int new_max_num_tokens=-1;
  int new_max_line_length=-1;
  int num_tokens, line_length;
  the_dictionary->num_items = 0;
  // count the number of items in the dictionary
  while(fgets(buffer, max_line_length, (FILE*) fp)) {
    line_length=strlen(buffer);
    char **columnX = NULL;
    num_tokens = getcols(buffer, ",", &columnX, max_num_tokens);
    if (num_tokens > new_max_num_tokens) 
      new_max_num_tokens = num_tokens;
    if (line_length > new_max_line_length)
      new_max_line_length = line_length;
    (*the_dictionary).num_items++;
    free(columnX);
  }
  fclose(fp);
  //printf("max line length %d  max num tokens %d num items %d\n",new_max_line_length,new_max_num_tokens,the_dictionary->num_items);
  // allocate the dictionary
  (*the_dictionary).theDictionary = (sentiment *) malloc(the_dictionary->num_items*sizeof(sentiment));
  int num_identities = readSentiments(the_dictionary, max_num_tokens, max_line_length);
  if (num_identities != the_dictionary->num_items) {
    return 0;
  } else {
    return num_identities;
  }
}

// function to read a sentiment dictionary (e.g. fbehaviours.dat)
// this text file should have the following format:
// each line is comma-separated, and contains a label, followed by
// three ratings (E,P,A) for males and three for females
// followed by a institutional filter string of 01 ****(which we ignore for now)
// the_dictionary must be big enough to hold all the entries
// some hard-coded constants in there that are ugly - 1024 max length of line and 255 max label length
int readSentiments(dictionary * a_dictionary, int max_num_tokens, int max_line_length)
{

  char **rows;
  int i,j,k,l;
  int num_tokens;
  // covenience names
  dict_type dtype = (*a_dictionary).dtype;
  char * fname = (*a_dictionary).fname;
  sentiment * the_dictionary = (*a_dictionary).theDictionary;
  int status;
  
  FILE * fp = fopen(fname,"r");
#if UNITY_WIN
  char *buffer = (char *)_alloca(sizeof(char) * max_line_length);
#else
  char buffer[max_line_length];
#endif
  if (!fp)
    return 0;
  rows = NULL; //getcols() requires it to be NULL. (Avoid dangling pointers, leaks e.t.c).

  k=0;
  while(fgets(buffer, max_line_length, (FILE*) fp)) {
    char **columnX = NULL;
    num_tokens = getcols(buffer, ",", &columnX, max_num_tokens);
    //the_dictionary[k].label=(char *) malloc(max_line_length*sizeof(char));
    strcpy(the_dictionary[k].label,columnX[0]);
    the_dictionary[k].epa = gsl_vector_alloc(SSD);
    for (j = 0; j < SSD; j++) 
      gsl_vector_set(the_dictionary[k].epa, j, strtod(columnX[ j+1 ],NULL));
    
    // hard-coded in here but this should really be read in from the dictionary as well
    // only used if dtype==MEAN
    the_dictionary[k].var = DTYPE_MEAN_VARIANCE;
    if (dtype == SD) {
      the_dictionary[k].variance = gsl_vector_alloc(SSD);
      for ( j = 0; j < SSD; j++)
	gsl_vector_set(the_dictionary[k].variance, j, strtod(columnX[ j + 4 ], NULL));
    } else if (dtype == COV) {
      the_dictionary[k].covariance = gsl_matrix_alloc(SSD,SSD);
      the_dictionary[k].cholesky = gsl_matrix_alloc(SSD,SSD);
      l=4;
      for ( i = 0; i < SSD; i++) {
	for ( j = 0; j < SSD; j++)
	  gsl_matrix_set(the_dictionary[k].covariance, i, j, strtod(columnX[ l++ ], NULL));
      }
      gsl_matrix_memcpy(the_dictionary[k].cholesky, the_dictionary[k].covariance);
      status =  gsl_linalg_cholesky_decomp1(the_dictionary[k].cholesky);
      if (status) // error - can't proceed
	return -1;
      //printMatrix(the_dictionary[k].covariance,the_dictionary[k].label,SSD,SSD);
    }
    // this could be true in general, but here I am using this so the institutional
    // filters are only read in (or attempted to be read in)
    // so this should be some kind of flag MEAN_INSTITUTIONS so you could have COV_INSTITUTIONS as well ...
    // right now, MEAN is the only one with institutional lists in any case.
    if (dtype == MEAN) {
      // a very specific format - 2 bits, blank 9 bits, blank, 3 bits all zeros or 1s
      char instlist[NUM_INSTITUTIONS+10];
      // skip leading whitespace
      char *p=columnX[ 7 ];
      while (isspace(* p)) p++;
      strcpy(instlist,p);
      if (strlen(instlist)<16) {
	fprintf(stderr,"ERROR in reading institution list\n");
	exit(1);
      }
      if (instlist[2]==' ' && instlist[12]==' ') {
	j=0;
	for (i = 0  ; i < 16; i++ ) {
	  if (i==2 || i==12) i++;
	  the_dictionary[k].the_institutions[j]= (int) instlist[i]-'0'; // strtol(instlist+i,(char **)NULL,1);
	  j++;
	}
      }
    }
    k++;
    free( columnX );
  }

  fclose(fp);
  return k;
}

// seek a dictionary called dname in the list dictionaries
// and return it, otherwise create a new dictionary if it does not exist,
// read from filename dname, add to the list dictionaries, increment num_dictionaries by 1
// and return a pointer to the dictionary (either new or old)
static dictionary * assign_dictionary(char * dname, char *dtypename, dictionary * dictionaries, int * num_dictionaries)
{
  dictionary * the_dictionary;
  dict_type dtype = lookUpDictType(dtypename);
  if ((the_dictionary = get_dictionary(dictionaries, *num_dictionaries, dname)) == NULL) {
    // read in new dictionary
#ifdef VERBOSE
    printf("reading in new dictionary %s\n",dname);
#endif
    if (!readSentimentDictionary(dtype,dname,dictionaries+(*num_dictionaries))) {
      printf("error reading dictionary %s\n",dname);
      exit(0);
    }
    the_dictionary = dictionaries+(*num_dictionaries);
    the_dictionary->gamma=DTYPE_MEAN_VARIANCE;  // may not be needed
    *num_dictionaries += 1;
  }
  return the_dictionary;
}

// assign all dictionaries based on file names in agents (NGT dictionaries agents[].entity_dicts_fname[])
// (to be found at path dpath)
// this function reads in all the dictionaries and sets up agent dictionary pointers to them
// output: dictionaries points to a list of num_dictionaries dictionary structures "dictionaries" these agents use
int assign_dictionaries(agentState *agents, int num_agents, dictionary * dictionaries, int *num_dictionaries)
{
  for (int j=0; j<num_agents; j++) {
    for (int i=0; i<NGT; i++) {
#ifdef VERBOSE
      printf("j:%d  i:%d  %s %s\n",j,i,agents[j].entity_dicts_fname[i],agents[j].entity_dict_types[i]);
#endif
      agents[j].entity_dicts[i]  = assign_dictionary(agents[j].entity_dicts_fname[i],
						     agents[j].entity_dict_types[i],
						     dictionaries, num_dictionaries);
      agents[j].entity_dicts[i]->dict_what=i;
#ifdef VERBOSE
      printf("%s num items in dictionary %s: %d\n",agents[j].name,
	     agents[j].entity_dicts[i]->fname,agents[j].entity_dicts[i]->num_items);
#endif
    }
    // the emotion dictionary
    agents[j].entity_dicts[EMOTION] = assign_dictionary(agents[j].entity_dicts_fname[EMOTION],
							agents[j].entity_dict_types[EMOTION],
							dictionaries, num_dictionaries);
#ifdef VERBOSE
    printf("%s num items in modifier dictionary %s: %d\n",agents[j].name,
	   agents[j].entity_dicts[EMOTION]->fname,agents[j].entity_dicts[EMOTION]->num_items);
#endif

  }
  return 1;
}

// ***************************************************************************************************************
// THE PARSER -- HOMEBREW EDITION 
// ***************************************************************************************************************
// reads in the simulation description
// afterwards, agents contain the agents (names, pointers to dictionaries, dynamics)
//
// input: sstate is the samplerState that will contain all the samples and parameters for each interaction
// input: fname is the input file (see example and notes in testsim.txt)
// output: agents are the agentState agents involved in the interaction
// output: num_agents is the number of agents involved
// output: dyads is a list of dyad structures, each corresponding to an interaction (a single agent-client dyad)
//         and giving an index into sstate->samples  and sstate->sentiParams, and an index for agent and client in that interaction
// output: num_dyads is the number of dyads 
// output: query_user is 0 (BAYESACT) if bayesact should be used for action selection,
//         1 (USER) if user should be queried in "interactive" mode
//         2 (EVENTS) if events file is used to specify a set of events to take
// output: max_iterations is the number of iterations to do (can be specified by the events file)
int readSimulationInput(char *fname, agentState * agents, int * num_agents,
			dyad *dyads, int *num_dyads, query_type * query_user, int * max_iterations, char ** events_filename)
{
  int i,k,j;
  static int max_num_tokens=100;  // must be bigger than expected
  static int max_line_length=2048;
  int num_agent_ids, num_client_ids;
  identity *agent_ids, *client_ids;
  int index, agent_index, client_index;
  int doing_interaction=0;
  int doing_agent=0;
  unsigned long rseed=0;
  c_entity_type the_entity;
  int the_institutions[NUM_INSTITUTIONS];
  // defaults
  (*num_dyads)=0;
  (*query_user)=BAYESACT;
  (*max_iterations)=0;
  strcpy(*events_filename,"");
  
  FILE * fp = fopen(fname,"r");
  char buffer[max_line_length];
  if (!fp) {
    fprintf(stderr,"file %s not found in readSimulationInput\n",fname);
    return -1;
  }
  
  *num_agents=0;
    
  int num_tokens, line_length;
  while(fgets(buffer, max_line_length, (FILE*) fp)) {
    if (!(buffer[0]=='/' && buffer[1]=='/')) {  // skip comments
      line_length=strlen(buffer);
      // remove trailing newline
      buffer[strcspn(buffer, "\n")] = 0;
      char **columnX = NULL;
      num_tokens = getcols(buffer, ":", &columnX, max_num_tokens);
      for (k=0; k<num_tokens; k++)
	trim(columnX[k]);
      //====================================================
      // start processing the input elements
      //====================================================
      if (strcmp(columnX[0],"num_iterations")==0)
	(*max_iterations) = atoi(columnX[1]);
      if (strcmp(columnX[0],"events")==0)
	strcpy(*events_filename,columnX[1]);
      if (strcmp(columnX[0],"simtype")==0) {
	//printf("got simtype as %s\n",columnX[1]);
	if (strcmp(columnX[1],"user") == 0)
	    (*query_user)=USER;
	if (strcmp(columnX[1],"bayesact") == 0)
	    (*query_user)=BAYESACT;
	if (strcmp(columnX[1],"events") == 0)
	    (*query_user)=EVENTS;
      }
      // done with an agent
      if (strcmp(columnX[0],"endagent")==0) {
	(*num_agents)++;
	doing_agent=0;
      }
      if (strcmp(columnX[0],"agent")==0) {
	// ========================================================================================================
	// read in dictionaries for identities, behaviours and modifiers, and dynamics for impressions and emotions
	// ========================================================================================================
#ifdef VERBOSE
	printf("new agent found with name %s\n",columnX[1]);
#endif
	strcpy(agents[*num_agents].name,columnX[1]);
	doing_agent=1;
	//default none
	agents[*num_agents].num_persona_ids=0;
	agents[*num_agents].alphaa=DEFAULT_ALPHA;
	agents[*num_agents].alphab=DEFAULT_ALPHA;
	agents[*num_agents].alphac=DEFAULT_ALPHA;
	agents[*num_agents].betaa=DEFAULT_BETA;
	agents[*num_agents].betac=DEFAULT_BETA;
	agents[*num_agents].deltaa=DEFAULT_AGENT_DELTA;
	agents[*num_agents].deltac=DEFAULT_AGENT_DELTA;
	agents[*num_agents].numsamples=DEFAULT_NUMSAMPLES;
	agents[*num_agents].do_resets=1;
	agents[*num_agents].default_action=0;
	agents[*num_agents].use_emotions=0;
	agents[*num_agents].emotion_planning=0;
	
	if (num_tokens > 2) {
	  // read in the persona for this agent
	  num_agent_ids = extract_ids(num_tokens-1,&agent_ids,columnX+1);
	  // copy to agent structure
	  if (num_agent_ids > 0) {
	    agents[*num_agents].persona_ids = (identity *) malloc(num_agent_ids*sizeof(identity));
	    agents[*num_agents].num_persona_ids = num_agent_ids;
	    memcpy(agents[*num_agents].persona_ids,agent_ids,num_agent_ids*sizeof(identity));
#ifdef VERBOSE
	    printf("num persona ids found for agent %s: %d\n",agents[*num_agents].name, agents[*num_agents].num_persona_ids);
#endif
	    free(agent_ids);
	  }
	}
      } else if (doing_agent) {
	if (strcmp(columnX[0],"dictionary")==0) {
	  the_entity = lookUpEntity(columnX[1]);
	  strcpy(agents[*num_agents].entity_dicts_fname[the_entity],DICT_PATH);
	  strcat(agents[*num_agents].entity_dicts_fname[the_entity],columnX[2]);
	  strcpy(agents[*num_agents].entity_dict_types[the_entity],columnX[3]);
#ifdef VERBOSE
	  printf("found agent dictionary for %s (index %d): %s with type %s\n",columnX[1],the_entity,agents[*num_agents].entity_dicts_fname[the_entity],agents[*num_agents].entity_dict_types[the_entity]);
#endif
	} else if (strcmp(columnX[0],"dynamics")==0) {
	  if (strcmp(columnX[1],"IMPRESSION")==0) {
#ifdef VERBOSE
	      printf("dynamics file: %s\n",columnX[2]);
#endif
	      strcpy(agents[*num_agents].dyn_fname,DICT_PATH);
	      strcat(agents[*num_agents].dyn_fname,columnX[2]);
	  } else if (strcmp(columnX[1],"EMOTION")==0) {
#ifdef VERBOSE
	    printf("emotions dynamics file: %s\n",columnX[2]);
#endif
	    strcpy(agents[*num_agents].mod_fname,DICT_PATH);
	    strcat(agents[*num_agents].mod_fname,columnX[2]);
	  }
	} else if (strcmp(columnX[0],"alphas")==0) {
#ifdef VERBOSE
	  printf("alphas: ");
	  for (i=1; i<num_tokens; i++) 
	    printf(" %s,",columnX[i]);
	  printf("\n");
#endif
	  if (num_tokens < 4) {
	    agents[*num_agents].alphaa=atof(columnX[1]);
	    agents[*num_agents].alphab=atof(columnX[1]);
	    agents[*num_agents].alphac=atof(columnX[1]);
	  } else {
	    agents[*num_agents].alphaa=atof(columnX[1]);
	    agents[*num_agents].alphab=atof(columnX[2]);
	    agents[*num_agents].alphac=atof(columnX[3]);
	  }
	} else if (strcmp(columnX[0],"betas")==0) {
#ifdef VERBOSE
	  printf("betas: ");
	  for (i=1; i<num_tokens; i++) 
	    printf(" %s,",columnX[i]);
	  printf("\n");
#endif
	  if (num_tokens < 3) {
	    agents[*num_agents].betaa=atof(columnX[1]);
	    agents[*num_agents].betac=atof(columnX[1]);
	  } else {
	    agents[*num_agents].betaa=atof(columnX[1]);
	    agents[*num_agents].betac=atof(columnX[2]);
	  }
	} else if (strcmp(columnX[0],"deltas")==0) {
#ifdef VERBOSE
	  printf("deltas: ");
	  for (i=1; i<num_tokens; i++) 
	    printf(" %s,",columnX[i]);
	  printf("\n");
#endif
	  if (num_tokens < 3) {
	    agents[*num_agents].deltaa=atof(columnX[1]);
	    agents[*num_agents].deltac=atof(columnX[1]);
	  } else {
	    agents[*num_agents].deltaa=atof(columnX[1]);
	    agents[*num_agents].deltac=atof(columnX[2]);
	  }
	} else if (strcmp(columnX[0],"flags")==0) {
#ifdef VERBOSE
	  printf("flags: ");
	  for (i=1; i<num_tokens; i++) 
	    printf(" %s,",columnX[i]);
	  printf("\n");
#endif
	  if (num_tokens >= 2) 
	    agents[*num_agents].do_resets=atof(columnX[1]);
	  if (num_tokens >= 3) 
	    agents[*num_agents].default_action=atof(columnX[2]);
	  if (num_tokens >= 4) 
	    agents[*num_agents].use_emotions=atof(columnX[3]);
	  if (num_tokens >= 5) 
	    agents[*num_agents].emotion_planning=atof(columnX[4]);
	} else if (strcmp(columnX[0],"numsamples")==0) {
#ifdef VERBOSE
	  printf("numsamples: ");
	  printf(" %s,",columnX[1]);
	  printf("\n");
#endif
	  if (num_tokens >=2) 
	    agents[*num_agents].numsamples=atof(columnX[1]);
	}
      }
      // ================================================================================================
      // set up intial agent and client identities for each interaction
      // ================================================================================================
      // we've finished one interaction so we set up the interaction and free up memory
      if (strcmp(columnX[0],"endinteraction")==0) {   // && (num_agent_ids > 0 || num_client_ids > 0))   {
	doing_interaction=0;

	// need to save this particular interaction consisting of
	// agent_index, client_index, agent_ids, num_agent_ids, client_ids, num_client_ids,
	// based on this, we can call setUpDyadicInteraction in the calling function
	dyads[*num_dyads].agent_index = agent_index;
	dyads[*num_dyads].client_index = client_index;

	if (num_agent_ids < 0 || num_client_ids < 0) { // no identity specified for this agent
#ifdef DO_INTERACT
	  fprintf(stderr,"Error: you MUST comment out the define of DO_INTERACT in fvarsample.h, \n\t because some agent has a null identity set\n\t OR give %s an identity for %s (and/or vice-versa)\n\t If this leaves you stuck, then you cannot use Interact actions and must change your sim file. Sorry.\n",agents[agent_index].name,agents[client_index].name);
	  exit(0);
#endif
	}
	// if there are no agent_ids specified, it will be initialized as a randomly sampled distribution
	// across the whole dictionary
	if (num_agent_ids > 0) {
	  dyads[*num_dyads].agent_ids = (identity *) malloc(num_agent_ids*sizeof(identity));
	  memcpy(dyads[*num_dyads].agent_ids,agent_ids,num_agent_ids*sizeof(identity));
	} else {
	  dyads[*num_dyads].agent_ids = NULL;
	}
	dyads[*num_dyads].num_agent_ids = num_agent_ids;

	if (num_client_ids > 0) {
	  dyads[*num_dyads].client_ids = (identity *) malloc(num_client_ids*sizeof(identity));
	  memcpy(dyads[*num_dyads].client_ids,client_ids,num_client_ids*sizeof(identity));
	} else {
	  dyads[*num_dyads].client_ids = NULL;
	}
	dyads[*num_dyads].num_client_ids = num_client_ids;

	//RANDOM SEEDS
	
	dyads[*num_dyads].random_seed = rseed;
	//#INSTITUTION STUFF
	// don't have to allocate its static at NUM_INSTITUTIONS
	//dyads[*num_dyads].institution_list=(int *) malloc(NUM_INSTITUTIONS*sizeof(int));
	//memcpy(dyads[*num_dyads].institution_list,the_institutions,NUM_INSTITUTIONS*sizeof(int));
	for (i=0; i<NUM_INSTITUTIONS; i++) 
	  dyads[*num_dyads].institution_list[i]=the_institutions[i];
	// increase the number of dyads found by 1
	(*num_dyads)++;
	
	if (num_client_ids>0) 
	  free(client_ids);
	if (num_agent_ids>0) 
	  free(agent_ids);
      }
      if (strcmp(columnX[0],"interaction")==0) {
	agent_index = find_agent(agents,*num_agents,columnX[1]);
	client_index = find_agent(agents,*num_agents,columnX[2]);
	if (agent_index<0 || client_index<0) {
	  printf("agents not found %s %s\n",columnX[1], columnX[2]);
	  exit(0);
	}
#ifdef VERBOSE
	printf("************************************** agents found %s %s %d %d\n",columnX[1], columnX[2],agent_index,client_index);
#endif
	doing_interaction=1;

	
	
      } else if (doing_interaction>=1) {
	// if its not an institution list first, this is handled in the else
	if (!(strcmp(columnX[0],"institution")==0) && !(strcmp(columnX[0],"rseed")==0)) {
	  // get the individual identity distributions
	  index=find_agent(agents,*num_agents,columnX[0]);
	  if (index==agent_index) {  // first item is self identity
	    num_agent_ids = extract_ids(num_tokens,&agent_ids,columnX);
#ifdef VERBOSE
	    printf("num agent ids: %d %d %s\n",num_agent_ids,agent_index,columnX[0]);
#endif
	  } else if (index==client_index) {
	    num_client_ids = extract_ids(num_tokens,&client_ids,columnX);
#ifdef VERBOSE
	    printf("num client ids: %d %d %s\n",num_client_ids,client_index,columnX[0]);
#endif
	  } else {
	    printf("agent not found %s\n",columnX[0]);
	    exit(0);
	  }
	} else if (strcmp(columnX[0],"rseed")==0) {
	  rseed=strtol(columnX[1],(char **)NULL,0);
	} else {
	  // if there are institutional filters, they must be specified right after the interaction
	  // read in a set of institutions
	  // institutional filters look like this: 11 010000000 000
	  // overt surmised lay business law politics academe medicine religion family sexual monadic group corporal
	  // NUM_INSTITUTIONS and INSTITUTION_LIST are defined in fvarsample.h
	  // INSTITUTIONS
	  char *the_institution;
	  int the_institution_count;
	  the_institution_count=0;
	  // to make sure the no specified institutions are 0 if none are specified (so they should all be 1 in this case)
	  for (k=0; k<NUM_INSTITUTIONS; k++) 
	    the_institutions[k]=0;
	  
	  k=1;
	  while(k<num_tokens) {
	    the_institution = columnX[k];
	    j=0;
	    while (j<NUM_INSTITUTIONS) {
	      if (strcmp(the_institution,INSTITUTION_LIST[j])==0) {
		the_institutions[j]=1;
		j=NUM_INSTITUTIONS+2;
	      }
	      if (j==NUM_INSTITUTIONS+2) 
		the_institution_count++;
	      j++;
	    }
	    k++; 
	  }
	  // to make sure the no specified institutions are 0 if none are specified (so they should all be 1 in this case)
	  if (the_institution_count==0) 
	    for (k=0; k<NUM_INSTITUTIONS; k++) 
	      the_institutions[k]=1;
	}
      }
      free(columnX);
    }
  }
  fclose(fp);
  return 1;

}

// Free dictionary as set by readSentimentDictionary
void freeSentimentDictionary(dictionary *  dict)
{
  for (int k=0; k < dict->num_items; k++) {
    if (dict->dtype == SD) {
      gsl_vector_free(dict->theDictionary[k].variance);
    } else if (dict->dtype == COV) {
      gsl_matrix_free(dict->theDictionary[k].covariance);
    }
  }
}

