/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.3.8)                              #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : April 2021                                                                            #
#  Copyright (C) 2017-2021  Jesse Hoey                                                           #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
# include file for the parser
#************************************************************************************************/
#ifndef __INPUTPARSER
#define __INPUTPARSER
#include "fvarsample.h"


#define DICT_PATH "../data/"
// must be bigger than expected
// number of tokens and line length per input file line
#define MAX_LINE_LENGTH 2048
#define MAX_NUM_TOKENS 100

//char INSTITUTION_LIST[NUM_INSTITUTIONS][10] = {"overt","surmised","lay","business","law","politics","academe","medicine","religion","family","sexual","monadic","group","corporal"};


typedef enum {
  NO_ACT=(-1),
  BAYES_ACT=(-2),
  INTER_ACT=(-3),
} action_type;

// setup
int setUpAgentsDyads(char *fname, agentState * agents, int * num_agents,
		     dyad *dyads, int *num_dyads, query_type * query_user, int * max_iterations, double gamma,
		     dictionary * dictionaries, int * num_dictionaries, event **events);

int assign_dictionaries(agentState *agents, int num_agents, dictionary * dictionaries, int *num_dictionaries);

// read events and input
int readEvents(char *events_fname, dyad *dyads, int num_dyads, agentState *agents, int num_agents, event **events);
int readSimulationInput(char *fname, agentState * agents, int * num_agents,
			dyad *dyads, int *num_dyads, query_type * query_user, int * max_iterations, char ** events_filename);


// institution filter list functions
void buildInstitutionFilterList(dictionary *the_dict, dyad *the_dyads, int num_dyads);
void freeInstitutionFilterList(dictionary *the_dict, int num_dyads);
void printInstitutionFilterList(dictionary *the_dict, dyad *the_dyads, int num_dyads);
void printInstitutionFilterListOneDyad(dictionary *the_dict, int k);

// reading/freeing sentiments and dictionaries
int readSentimentDictionary(dict_type dtype, char * filename, dictionary * the_dictionary);
int readSentiments(dictionary * a_dictionary, int max_num_tokens, int max_line_length);
void freeSentimentDictionary(dictionary * the_dictionary);

#endif

