/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.2)                                #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2021                                                                            #
#  Copyright (C) 2021  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- stub version
#************************************************************************************************/


#include "institutionalAPP.h"


// initialize and free

void initializeXXObs(xostate **xobs)
{
  // need to have this to maintain consistency, but it is not used
  *xobs = (xostate *) malloc(sizeof(xostate));
  (*xobs)->dummy = 0;
}
void freeXXObs(xostate **xobs)
{
}
// update any additional xstate based on xstate
// and on the actions taken by the two agents
// which are the indices of the denotative actions taken by agent and client
// in the dictionaries specified as BEHAVIOUR in the input file
// generally this is a draw from P(x'|x'_{ba},x'_{bc})
void updateXXSample(gsl_rng *r, xxstate *next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour)
{
  // a stub example
  return;
}

// new versions that takke agentState instead of fixed number of behaviours
void updateSampleXbc(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (o_bc >= 0) {
    (*x_bc)=o_bc;
  } else {
    // irrelevant for getAction - really this plays no role anywhere other than as the expectation of what the client will do,
    // which will be filled in anyways 
    //filter by institution here!
    // something like drawRandomFiltered(the_agent->entity_dicts[BEHAVIOUR],the_agent->institution_dicts[BEHAVIOUR]);
    // this will select a random element from entity_dicts that has a flag set in institution_dicts
    // the flag has a number which in binary is the mask for the institution list.

    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_bc) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];

    //int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_bc) = gsl_rng_uniform_int(r, num_behaviours);


    
  }
}
void updateSampleXba(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (a >= 0) {
    (*x_ba)=a;
  } else {
    // want to draw a sample from the denotative distribution over actions
    // here is code to exclude "discipline", but randomly choose otherwise
    /*
      int done=0;
    while (!done) {
      (*x_ba) = gsl_rng_uniform_int(r, num_behaviours);
      if (!(strcmp(the_agent->entity_dicts[BEHAVIOUR]->theDictionary[(*x_ba)].label,"discipline")==0))
	done=1;
    }
    */
    //    FILTER BY INSTITUTION HERE!
    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_ba) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];

    //int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_ba) = gsl_rng_uniform_int(r, num_behaviours);

    
    // here is code to include only two behaviours and choose randomly
    /*
    int tmpact = gsl_rng_uniform_int(r, 2);
    if (tmpact) 
      (*x_ba)=0; // abandon
    else
      (*x_ba)=248; // ignore
    // entertain other possibilities
    int tmp=gsl_rng_uniform_int(r, 4);
    tmp=tmp+1;
    // this doesn't work with two agents that are the same because they both switch 
    if (tmp==1) {
      *x_ba=373;  // rape
    }
    if (tmp==2) {
      *x_ba=159;  //dress
    }
    if (tmp==3) {
      *x_ba=424;  // sit next to
    }
    if (tmp==4) {
      *x_ba=489;  // watch
    }
    */
  }
}

// institutional filters look like this: 11 010000000 000
// over surmised
// Lay business Law Politics Academe Medicine Religion Family Sexual
// Monadic Group Corporal 
//
//unsigned int institution[14];

void updateSampleXac(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int client_behaviour, dyad *the_dyad, agentState *the_agent)
{
  // filter by institution here! 
  //int num_identities=the_agent->entity_dicts[AGENT]->num_items;


  int num_ic_identities = the_agent->entity_dicts[AGENT]->num_ic_items[the_dyad->dyad_index];
  int tmp_id=gsl_rng_uniform_int(r,num_ic_identities);
  (*x_ap) = the_agent->entity_dicts[AGENT]->list_ic_converter[the_dyad->dyad_index][tmp_id];

  num_ic_identities = the_agent->entity_dicts[CLIENT]->num_ic_items[the_dyad->dyad_index];
  tmp_id=gsl_rng_uniform_int(r,num_ic_identities);
  (*x_cp) = the_agent->entity_dicts[CLIENT]->list_ic_converter[the_dyad->dyad_index][tmp_id];

  return;
}


// this isa a draw from P(x'_{bc}|x_x, x_bc, x_ba, x_a, x_c) where x_x is in current_xxstate and
// x_bc, x_ba, x_a, x_c are in current_xstate
// this will normally be Pr(x_{bc}'|x,x_{bc}) 
// o_bc is included because you might want to sample from Pr(x_{bc}'|x,x_{bc},o_{bc}') or at least use
// a proposal distribution that is based on the observation o_bc
void updateXbcSample(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, int num_behaviours)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (o_bc >= 0) {
    (*x_bc)=o_bc;
  } else {
    // here is a version that choose randomly - this is normally what we should do but it wastes samples
    // here num_behaviours is the size of the behaviour dictionary being used - would need to make that accessible
    // here somehow
    // in general, I should (1) pass in the_agent instead of num_behaviours
    // (2) implement a function here that maps from the identities of agent and client in current_xstate,
    // along with the value of xx in current_xxstate, and gives a random draw from a denotative distribution
    // that represents the posterior over x_bc
    //    if (current_xstate[0]==295)
    (*x_bc) = gsl_rng_uniform_int(r, num_behaviours); 
    //(*x_bc)  = 373; // gsl_rng_uniform_int(r, 1);
  }
}
// this is a draw from P(x'_{ba}|x_x,x_bc, a, x_a, x_c)
void updateXbaSample(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, int num_behaviours)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (a >= 0) {
    (*x_ba)=a;
  } else {
    // here is a version that choose randomly - this is normally what we should do but it wastes samples
    // here num_behaviours is the size of the behaviour dictionary being used - would need to make that accessible
    // here somehow
    (*x_ba) =  gsl_rng_uniform_int(r, num_behaviours);
    //  set institutional filters here:
    // (*x_ba) = 18;  to include behaviour 18, etc ...;
    // pick randomly from first two - just for testing
    //(*x_ba)=373; // gsl_rng_uniform_int(r, 1);
    
  }
}
// this is Pr(xa',xc'|xa,xc,agent_beh,client_beh)
// not implemented 
void updateXacSample(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int client_behaviour)
{
  return;
}

// this is R(x)
double getXXUtility(xxstate * xx_state)
{
  return 0.0;
}


// weight x' based on o_x'
// generally this returns P(o_x | x')
double weightXXSample(gsl_rng *r, xxstate * xx_state, xostate * xx_observation)
{
  return 1.0;
}
// weight the x_bc sample according to the observation - by default it should be identical
// however, there could be some observation noise that comes into play here. It would also have
// to be present in the weighting of the XX sample though
// weight x_{bc}' based on \omega_{x_{bc}} - this is P(o'_bc | x'_bc)
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc)
{
  if ((*x_bc)==o_bc)
    return 1.0;
  return 0.0;
}
// draw a sample from P(o_x | x) 
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation)
{
  xx_observation = NULL;
}
// draw a sample from P(o_bc | x_bc) 
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc)
{
  // could add noise here
  (*o_bc) = x_bc;
}
//----------------------------------------------------------------------------------------------------
// other helper functions
//----------------------------------------------------------------------------------------------------


// function convert a string to an xx_observation (for user input)
int stringToXXObs(char * str, xostate *xx_observation)
{
  return 1;  // success
}
void xxObsToString(char * retstr,xostate *xx_observation)
{
  strcpy(retstr,"");
}
int compareXXObs(xostate * xx1, xostate *xx2)
{
  return 1;
}

// function to print the xx state and observatoin
void printRawXXstate(xxstate * xx)
{
  return;
}
void printRawXXObs(xostate * xx_observation)
{
  return;
}

// functions to initialize, add, divide and copy XX states
void initializeXXstate(xxstate *xx)
{
  return;
}

void addToXXstate(xxstate *xx_avg, xxstate *xx, double weight)
{
  return;
}
void divideXXstate(xxstate * xx,double v)
{
  return;
}
void copyXXstate(xxstate *dest, xxstate *src)
{
  return;
}

