/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- for Prisoner's dilemma
# which doesn't actually use the X observations but its included here for completeness
#************************************************************************************************/

#include "pdAPP.h"

// initialize and free
void initializeXXstate(xxstate *xx)
{
  setXXstate(xx,0,0);
}
void setXXstate(xxstate *xx, int amount_won, int whose_turn)
{
  xx->amount_won=amount_won;
  xx->whose_turn=whose_turn;  // should be settable
  xx->other_last_beh=-1;
}
void initializeXXObs(xostate **xobs)
{
  *xobs = (xostate *) malloc(sizeof(xostate));
  (*xobs)->obs_amount_won = 0;
}
void freeXXObs(xostate **xobs)
{
  free(*xobs);
}
// update any additional xxstate based on xxstate
// and on the actions taken by the two agents
// which are the indices of the denotative actions taken by agent and client
// in the dictionaries specified as BEHAVIOUR in the input file
// in general this is a draw from P(x'|x'_{ba},x'_{bc})
void updateXXSample(gsl_rng *r, xxstate *next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour)
{
  
  // 0 = give, 1 = take  --> note: this mapping is defined implicitly by the dictionary being used -- this is awkward!

  // give 2/ take 1
  /*
  next_xxstate->amount_won = 0;
  if (agent_behaviour==1)
    next_xxstate->amount_won += 1;
  if (client_behaviour == 0)
    next_xxstate->amount_won += 2;
  */
  // in general
  int T=3;   // temptation
  int C=2;   //cooperation
  int D=1;   // defection
  int S=0;  // sucker
  // for agent that goes first:
  next_xxstate->whose_turn = ( 1 - current_xxstate->whose_turn );
  next_xxstate->amount_won = 0;
  if (agent_behaviour==NULL && client_behaviour==NULL) 
      return;
  //printf("agent behaviour %d whose turn %d other_last_beh %d client behaviour %d\n",agent_behaviour,current_xxstate->whose_turn,current_xxstate->other_last_beh,client_behaviour);
  if (current_xxstate->whose_turn==0) {
    // first turn for aria
    if (current_xxstate->other_last_beh<0){
      next_xxstate->amount_won = 0;
      next_xxstate->other_last_beh=agent_behaviour;
    } else {
      // second turn for player  - compare what player did this turn (agent_behaviour) with what aria did last turn (stored in other_last_beh)
      // aria gave
      if (current_xxstate->other_last_beh==0) {
	// player gave
	if (agent_behaviour==0) 
	  next_xxstate->amount_won = C;
	else if (agent_behaviour==1)
	  next_xxstate->amount_won = T;
      } else if (current_xxstate->other_last_beh==1) {
	if (agent_behaviour==0)
	  next_xxstate->amount_won = S;
	else if (agent_behaviour==1)
	  next_xxstate->amount_won = D;
      }
      next_xxstate->other_last_beh = (-1);
    }
  } else {
    // first turn for player
    if (current_xxstate->other_last_beh<0){
      next_xxstate->amount_won = 0;
      next_xxstate->other_last_beh=client_behaviour;
    } else {
      // second turn for aria - compare what aria did last turn (stored in other_last_beh) to what the client did this turn
      // aria gave
      if (current_xxstate->other_last_beh==0) {
	// player gave
	if (client_behaviour==0) 
	  next_xxstate->amount_won = C;
	else if (client_behaviour==1)
	  next_xxstate->amount_won = S;
      } else if (current_xxstate->other_last_beh==1) {
	if (client_behaviour==0)
	  next_xxstate->amount_won = T;
	else if (client_behaviour==1)
	  next_xxstate->amount_won = D;
      }
      next_xxstate->other_last_beh = (-1);
    }
    
  }      
  //next_xxstate->whose_turn = ( 1 - current_xxstate->whose_turn );
  //printf("amount won: %d\n",next_xxstate->amount_won);
  //if (gsl_rng_uniform(r) > 0.9)   // can add some randomness
  //    next_xxstate->amount_won = 0;
  return;
}

// this is Pr(x_{bc}'|x,x_{bc})
// this is a draw from P(x'_{bc}|x_x, x_bc, x_ba, x_a, x_c) where x_x is in current_xxstate and
// was : void updateXbcSample(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, int num_behaviours)
void updateSampleXbc(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly the same
  if (o_bc >= 0) {
    // cheat
    (*x_bc)=o_bc;
  } else {
    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_bc) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];
  }
}
// this is Pr(x_{ba}'|x,x_{ba})
void updateSampleXba(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (a >= 0) {
    (*x_ba)=a;
  } else {
    // want to draw a sample from the denotative distribution over actions
    //    FILTER BY INSTITUTION HERE!
    //    int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_ba) = gsl_rng_uniform_int(r, num_behaviours);
    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_ba) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];
    
  }
}
// this is Pr(xa',xc'|xa,xc,agent_beh,client_beh)
void updateSampleXac(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int client_behaviour, dyad *the_dyad, agentState *the_agent)
{
  double pxa,pxc;
  // default copy over
  *x_ap = x_a;
  *x_cp = x_c;
  double AGENT_DELTA=the_agent->deltaa;  //DEFAULT_AGENT_DELTA;
  double CLIENT_DELTA=the_agent->deltac;  //DEFAULT_AGENT_DELTA;
  
  //this value should be a property of the_agent once the ssstate becomes part of the dyad
  // or is accessible through the dyad
  // This is part of an agent's delta value.
  // the other part is the DISRUPT_NOISE
  
  double theid[SSD];
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[AGENT]->theDictionary[x_a].epa,i)+
      gsl_ran_gaussian(r,AGENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_ap, theid, the_agent->entity_dicts[AGENT]);
  
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[CLIENT]->theDictionary[x_c].epa,i)+
      gsl_ran_gaussian(r,CLIENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_cp, theid, the_agent->entity_dicts[CLIENT]);


  return;
}

// this is R(x)
double getXXUtility(xxstate * xx_state)
{
  return xx_state->amount_won;
}

// weight x' based on o_x'
// generally this returns P(o_x | x')
double weightXXSample(gsl_rng * r, xxstate * xx_state, xostate * xx_observation)
{
  if (xx_state->amount_won == xx_observation->obs_amount_won)
    return 1.0-PERCEPTION_NOISE;

  return PERCEPTION_NOISE;
}
// weight x_{bc}' based on \omega_{x_{bc}} - this is P(o'_bc | x'_bc)
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc)
{
  // could be somewhat uncertain about what this agent perceives
  if ((*x_bc)==o_bc)
    return 1-PERCEPTION_NOISE;
  return PERCEPTION_NOISE;
}
// draw a sample from P(o_x | x) 
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation)
{
  // could add noise here
  (*xx_observation).obs_amount_won = xx_state->amount_won;
}
// draw a sample from P(o_bc | x_bc) 
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc)
{
  // could add noise here
  (*o_bc) = x_bc;
}

// other helper functions
int stringToXXObs(char * str, xostate *xx_observation)
{
  xx_observation->obs_amount_won = atoi(str);
  return 1;  // success  -1 for failure
}
void xxObsToString(char * retstr, xostate *xx_observation)
{
  sprintf(retstr,"%d",xx_observation->obs_amount_won);
}

void copyXXstate(xxstate *dest, xxstate *src)
{
  dest->amount_won = src->amount_won;
  dest->whose_turn = src->whose_turn;
  dest->other_last_beh = src->other_last_beh;
}

void printRawXXstate(xxstate * xx) {
  printf("amount won: %d\n",xx->amount_won);
  printf("whose turn %d\n",xx->whose_turn);
  printf("others last behaviour: %d\n",xx->other_last_beh);
}
void printRawXXObs(xostate * xx_observation)
{
  printf("amount won observed %d\n",xx_observation->obs_amount_won);
}
int compareXXObs(xostate * xx1, xostate *xx2)
{
  return xx1->obs_amount_won == xx2->obs_amount_won;
}
void addToXXstate(xxstate *xx1, xxstate *xx2, double weight)
{
  xx1->amount_won += xx2->amount_won*weight;
}
void divideXXstate(xxstate * xx,double v)
{
  xx->amount_won  = (int) ((double) xx->amount_won)/v;
}


