/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- for Prisoner's dilemma
#  this one is to use with the new BayesACT 2.3.8
# also - this one "requires" but doesn't enforce turn-taking -
# and it always goes agent first, then client, then reward after client action
#************************************************************************************************/



#include "pdbayesactAPP.h"

// initialize and free
void initializeXXObs(xostate **xobs)
{
  *xobs = (xostate *) malloc(sizeof(xostate));
  (*xobs)->obs_amount_won = 0;
}
void freeXXObs(xostate **xobs)
{
  free(*xobs);
}
// update any additional xxstate based on xxstate
// and on the actions taken by the two agents
// which are the indices of the denotative actions taken by agent and client
// in the dictionaries specified as BEHAVIOUR in the input file
// in general this is a draw from P(x'|x'_{ba},x'_{bc})
void updateXXSample(gsl_rng *r, xxstate *next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour)
{
  
  // 0 = give, 1 = take  --> note: this mapping is defined implicitly by the dictionary being used -- this is awkward!

  // give 2/ take 1
  /*
  next_xxstate->amount_won = 0;
  if (agent_behaviour==1)
    next_xxstate->amount_won += 1;
  if (client_behaviour == 0)
    next_xxstate->amount_won += 2;
  */
  
  
  // in general
  int T=3;   // temptation
  int C=2;   //cooperation
  int D=1;   // defection
  int S=0;  // sucker
  if (current_xxstate->turn == 1) {
    // this means the *agent acted on the last iteration (was the "actor"), while *client acted on this turn (is therefore the "agent")
    // so this looks really awkward, but it works - 
    if (current_xxstate->xbasave==0) {
      if (agent_behaviour==0) 
	next_xxstate->amount_won = C;
      else if (agent_behaviour==1)
	next_xxstate->amount_won = S;
    } else if (current_xxstate->xbasave==1) {
      if (agent_behaviour==0)
	next_xxstate->amount_won = T;
      else if (agent_behaviour==1)
	next_xxstate->amount_won = D;
    }
    next_xxstate->turn=0;
    next_xxstate->xbasave = current_xxstate->xbasave;
  } else {
    next_xxstate->turn=1;
    next_xxstate->amount_won=0.0;
    next_xxstate->xbasave = agent_behaviour;
  }
    
      
  //  printf("amount won: %d\n",next_xxstate->amount_won);
  //if (gsl_rng_uniform(r) > 0.9)   // can add some randomness
  //    next_xxstate->amount_won = 0;
  return;
}

// new versions that takke agentState instead of fixed number of behaviours
// this is a draw from P(x'_{bc}|x_x, x_bc, x_ba, x_a, x_c) where x_x is in current_xxstate and
void updateSampleXbc(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (o_bc >= 0) {
    (*x_bc)=o_bc;
  } else {
    // irrelevant for getAction - really this plays no role anywhere other than as the expectation of what the client will do,
    // which will be filled in anyways 
    //filter by institution here!
    // something like drawRandomFiltered(the_agent->entity_dicts[BEHAVIOUR],the_agent->institution_dicts[BEHAVIOUR]);
    // this will select a random element from entity_dicts that has a flag set in institution_dicts
    // the flag has a number which in binary is the mask for the institution list.
    //int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_bc) = gsl_rng_uniform_int(r, num_behaviours);

    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_bc) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];

    // in pdAPP.c see the other funky stuff I do here. Instead of predicting the agent's behaviour based only on the institution,
    // I base it on the denotative identities currently in play. e.g. if agent and client are friends, its very likely client will
    // collaborate. I could also put this in below in updateSampleXba for the agent's own action (computed in getAction), which I could base
    // on the current denotative identities. However, note that these are all randomized draws across the samples, so all we are doing by
    // fiddling with Xbc and Xba is modifying what gets sampled. This is not necessary unless things get trickier I think.
    
  }
}
// this is Pr(x_{ba}'|x,x_{ba})
void updateSampleXba(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, dyad *the_dyad, agentState *the_agent)
{
  // client_behaviour is actually an observation but by default its exactly  the same
  // usually we should draw a sample here *independently* of o_bc - it gets used to generate a weight later
  // with a call to weightXXbcSample (see below)
  if (a >= 0) {
    (*x_ba)=a;
  } else {
    // want to draw a sample from the denotative distribution over actions
    //    FILTER BY INSTITUTION HERE!
    //    int num_behaviours=the_agent->entity_dicts[BEHAVIOUR]->num_items;
    //(*x_ba) = gsl_rng_uniform_int(r, num_behaviours);
    int num_ic_behaviours = the_agent->entity_dicts[BEHAVIOUR]->num_ic_items[the_dyad->dyad_index];
    int tmp_id=gsl_rng_uniform_int(r,num_ic_behaviours);
    (*x_ba) = the_agent->entity_dicts[BEHAVIOUR]->list_ic_converter[the_dyad->dyad_index][tmp_id];
    
  }
}
// this is Pr(xa',xc'|xa,xc,agent_beh,client_beh)
void updateSampleXac(gsl_rng *r, int * x_ap,  int x_a, int * x_cp, int x_c, int agent_behaviour, int client_behaviour, dyad *the_dyad, agentState *the_agent)
{
  // the identity stays the same by default
  (*x_ap) = x_a;
  (*x_cp) = x_c;

  double AGENT_DELTA=DEFAULT_AGENT_DELTA;
  //this value should be a property of the_agent once the ssstate becomes part of the dyad
  // or is accessible through the dyad
  // I think this may be the agent's delta value. I'm not really sure why.
  // possibly match this with the DISRUPT_NOISE as delta
  // add some random noise to these in EPA space
  double theid[SSD];
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[AGENT]->theDictionary[x_a].epa,i)+
      gsl_ran_gaussian(r,AGENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_ap, theid, the_agent->entity_dicts[AGENT]);
  
  for (int i=0; i<SSD; i++) 
    theid[i]=gsl_vector_get(the_agent->entity_dicts[CLIENT]->theDictionary[x_c].epa,i)+
      gsl_ran_gaussian(r,AGENT_DELTA);
  somaticTransform(DENOTATIVE, r, the_dyad->dyad_index, x_cp, theid, the_agent->entity_dicts[CLIENT]);


  return;
}


// this is R(x)
double getXXUtility(xxstate * xx_state)
{
  if (xx_state->turn == 1) 
    return xx_state->amount_won;
  else
    return 0.0;
}

// weight x' based on o_x'
// generally this returns P(o_x | x')
double weightXXSample(gsl_rng * r, xxstate * xx_state, xostate * xx_observation)
{
  
  if (xx_state->turn==1)
    return 1.0;
  if (xx_state->amount_won == xx_observation->obs_amount_won)
    return 1.0-PERCEPTION_NOISE;
  else
    return PERCEPTION_NOISE;
}
// weight x_{bc}' based on \omega_{x_{bc}} - this is P(o'_bc | x'_bc)
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc)
{
  // could be somewhat uncertain about what this agent perceives
  if ((*x_bc)==o_bc)
    return 1-PERCEPTION_NOISE;
  return PERCEPTION_NOISE;
}
// draw a sample from P(o_x | x) 
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation)
{
  // could add noise here
  (*xx_observation).obs_amount_won = xx_state->amount_won;
}
// draw a sample from P(o_bc | x_bc) 
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc)
{
  // could add noise here
  (*o_bc) = x_bc;
}

// other helper functions
// this function is what converts a user input string str giving xx_observations and converts to an observation for this guy
int stringToXXObs(char * str, xostate *xx_observation)
{
  xx_observation->obs_amount_won = atoi(str);
  return 1;  // success  -1 for failure
}
void xxObsToString(char * retstr, xostate *xx_observation)
{
  sprintf(retstr,"%d",xx_observation->obs_amount_won);
}

void copyXXstate(xxstate *dest, xxstate *src)
{
  dest->amount_won = src->amount_won;
  dest->turn = src->turn;
  dest->xbasave = src->xbasave;
}

void printRawXXstate(xxstate * xx) {
  printf("amount won: %d\n",xx->amount_won);
}
void printRawXXObs(xostate * xx_observation)
{
  printf("amount won observed %d\n",xx_observation->obs_amount_won);
}
int compareXXObs(xostate * xx1, xostate *xx2)
{
  return (xx1->obs_amount_won == xx2->obs_amount_won);
}
void initializeXXstate(xxstate *xx)
{
  xx->amount_won=0;
  xx->turn=0;
  xx->xbasave=0;
}
void addToXXstate(xxstate *xx1, xxstate *xx2, double weight)
{
  xx1->amount_won += xx2->amount_won*weight;
  xx1->turn = xx2->turn;
  xx1->xbasave = xx2->xbasave;
}
void divideXXstate(xxstate * xx,double v)
{
  xx->amount_won  = (int) ((double) xx->amount_won)/v;
}


