/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2.1)                                #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2021                                                                            #
#  Copyright (C) 2021  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  testfvarsample
#************************************************************************************************/
#include <stdio.h>
#include <sys/time.h>
#include "fvarsample.h"
#include "inputparser.h"

#define WEIGHT_THRESHOLD 10e-300
int main(int argc, char **argv)
{
  double alpha=1.0;
  double beta = 0.0001;
  double gamma=0.1;
  char *dyn_fname = "tdynamics-male.dat";
  int num_samples=1000000;

  printf("starting BayesACT\n");

  if (argc > 1) 
    alpha = atof(argv[1]);
  if (argc > 2) 
    beta = atof(argv[2]);
  if (argc > 3) 
    gamma = atof(argv[3]);
  double betaa=beta;
  double betac=beta;
  
  samplerState *sstate = (samplerState *) malloc(sizeof(samplerState));
  sstate->current_num_interactions = 0;
  sstate->sentiParams = (sentimentParams *) malloc(MAX_NUM_INTERACTIONS*sizeof(struct sentimentParameters));
  for (int i=0; i<MAX_NUM_INTERACTIONS; i++) {
    (sstate->sentiParams+i)->is_allocated = 0;
  }
  sstate->alphaa = alpha;
  sstate->alphab = alpha;
  sstate->alphac = alpha;
  sstate->betaa=betaa;
  sstate->betac=betac;

  // set up stuff for random number generation
  gsl_rng_env_setup();
  unsigned long int the_seed = 0;
#ifdef RNG_SEED
  the_seed = RNG_SEED; // will seed with 0 by default
  gsl_rng_default_seed = RNG_SEED;
#endif

  sstate->T = gsl_rng_default;  // change the random number generator here
  sstate->r = gsl_rng_alloc (sstate->T);
  
  // re-seed if this is specified
  struct timeval start;
  gettimeofday(&start, NULL);
#ifndef RNG_SEED
  the_seed = start.tv_usec;
  printf("seeding with %lu\n",the_seed);
  gsl_rng_set(sstate->r, the_seed);
#endif
  

  int dyad_index = 0;
  if (setupParameters(sstate, dyn_fname, NULL)) 
    return -1;
    

  double * f_out, * tau_out;

  f_out = (double *) malloc(VD*sizeof(double));
  tau_out = (double *) malloc(VD*sizeof(double));
  //fundamental sentiment to start - behaviour is not relevant
  // secretary dress boss
  //double f[VD] = {1.03, 0.01, -0.18, 0.95, 0.07, -0.33, 0.48, 2.16, 0.94}; // secretary dress boss
  //double f[VD] = {-1.89, -0.23, 0.89, 0, 0, 0, -1.83, -1.02, 1.03}; // sexist X slut
  //double f[VD] = {1.51, 1.45, -0.18, 0,0,0, 1.49, 0.31, 0.75}; // tutor X student
  //double f[VD] = {2.4,1.65,-0.33,0,0,0,1.8,0.46,0.08}; // tutor X student  USA 2014
  double f[VD]= {1.4446,0.9965,0.5312,0,0,0,0.7782,0.0054,0.4714}; // citizen X coal miner

  //miner,     0.8814,-0.0354,0.4874,1.0626,0.5409,0.2053,0.5409,3.0246,1.2113,0.2053,1.2113,2.1013
  //coal miner,0.7782,0.0054,0.4714,1.8904,0.4388,1.0839,0.4388,3.0720,0.4168,1.0839,0.4168,3.1888
  //accuse,-1.5648,1.2459,1.3676,1.6624,-0.2161,-0.4417,-0.2161,2.8950,0.0626,-0.4417,0.0626,1.7263

  //double f[VD] = {1.51, 1.45, -0.18, 2.76, 2.39, 0.9, 1.49, 0.31, 0.75}; // tutor teach student
  //double f[VD] = {1.49, 0.31, 0.75,  0.58, -0.52, -0.95, 1.51, 1.45, -0.18}; // student obey tutor
  //double f[VD] = {1.49, 0.31, 0.75,  -1.05, 0.39, 1.69, 1.51, 1.45, -0.18}; // student yell at tutor
  //double f[VD] = {1.45, -0.76, 2.1,-3.11, -0.09, 1.11, 2.46, 2.54, 0.76}; // child stab father
  //double f[VD] = {1.45, -0.76, 2.1,2.68, 2.11, 1.26, 2.46, 2.54, 0.76}; // child kiss father
  //double f[VD] = {1.03,   0.01,  -0.18, 1.01, 0.05, 0.02,     1.26,   1.93,   1.44};
  //double f[9] = {1.03,   0.01,  -0.18, 0.58, -0.52, -0.95, 1.26,   1.93,   1.44};  //obey
  //tau is a direct copy of f to start with
  double * tau = (double *) malloc(VD*sizeof(double));

  //double behaviour_observation[3]={0.76,1.54,0.26};  //exonerate
  //double behaviour_observation[3]={0.58, -0.52, -0.95};   //obey
  //double behaviour_observation[3]={1.01, 0.05, 0.02};  //pet 
  //double behaviour_observation[3]={0.95, 0.07, -0.33};//dress
  //double behaviour_observation[3] = {2.76, 2.39, 0.9}; // teach
  //double behaviour_observation[3] = {-1.14,-0.26,1.07}; // pinch
  //double behaviour_observation[3] = {-1.05, 0.39, 1.69}; // yell at
  //double behaviour_observation[3]={ -2.89, -0.82, -1.57}; // steal from,
  //double behaviour_observation[3]={ -3.1, -0.59, 1.07}; // rape
  //double behaviour_observation[3] = {-3.11, -0.09, 1.11}; // stab  Indiana 2005
  double behaviour_observation[3]={-1.5648,1.2459,1.3676}; // accuse
  //double behaviour_observation[3]={-3.5238,0.7500,0.1831}; //degrade
  //double behaviour_observation[3] = {-3.11, -0.09, 1.11}; // stab
  //double behaviour_observation[3] = {2.68, 2.11, 1.26}; // kiss
  memcpy(tau,f,VD*sizeof(double));

  printf("fundamental input: ");
  for (int i=0; i<VD; i++)
    printf("%g ",f[i]);
  printf("\n");
  printf("transient input: ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau[i]);
  printf("\n");

  double defl = 0.0;
  // Draws a SINGLE sample? I think so
  // this is the core P(f'|f,tau) function
  /*
  sampleNextSentimentRaw(sstate, dyad_index, BEH_AGENT, RESET_NONE, f, tau, f_out, tau_out);
  


  printf("fundamental output: ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_out[i]);
  printf("\n");
  printf("transient output: ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_out[i]);
  printf("\n");
  for (int i=0; i<VD; i++)
    defl += pow(f_out[i]-tau_out[i],2);
  printf("deflection (bayesact): %g\n",defl);
      
  */
  // ***********************************************************************************************
  // Computations involving interact equations directly
  // ***********************************************************************************************
  double tau_out_interact[VD];
  double tau_out_interact2[VD];
  double tau_interact[VD];
  double f_interact[VD];
  for (int i=0; i<VD; i++)
    f_interact[i]=f[i];
  for (int i=0; i<SSD; i++) 
    f_interact[i+SSD]=behaviour_observation[i];
  double f_reset[VD];
  double tau_reset[VD];
  
  memcpy(tau_interact,f_interact,VD*sizeof(double));
  
  printf("fundamental (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_interact[i]);
  printf("\n");

  printf("Interact optimal identites and behaviour (NOW): ");
  computeInteractMean(sstate, BEH_AGENT, f_interact,tau_interact);
  printInteractMean(sstate);

  // function computes the transients in tau_out from the f_b and tau_a/tau_c
  defl  = formImpression(sstate, BEH_AGENT, f_interact, tau_interact, tau_out_interact);


    // this should be identical to that produced by interact for the event 
  printf("transient output (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_out_interact[i]);
  printf("\n");
  printf("deflection (interact): %g\n",defl);
  printf("fundamental (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_interact[i]);
  printf("\n");

  printf("Interact optimal identities and behaviour (NEXT): ");
  //computeInteractMean(sstate,  BEH_AGENT, f_interact,tau_out_interact);
  computeInteractMean(sstate, BEH_AGENT, f_interact,tau_out_interact);
  printInteractMean(sstate);


  // function computes the transients in tau_out from the f_b and tau_a/tau_c
  defl  = formImpression(sstate, BEH_AGENT, f_interact, tau_out_interact, tau_out_interact2);
  printf("transient output (interact): ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_out_interact2[i]);
  printf("\n");
  printf("deflection (interact): %g\n",defl);
  

  // ***********************************************************************************************
  // Computations involving BayesACT
  // ***********************************************************************************************

  double * f_avg, * tau_avg;

  f_avg = (double *) malloc(VD*sizeof(double));
  tau_avg = (double *) malloc(VD*sizeof(double));

  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  double theweight,totalweight;
  totalweight=0.0;

  // first compute without the observation to get the distribution over f_b
  // this relies on sampleNextSentimentRaw being called to set up the matrices -
  // but this sample will do it without f_b in the prior so that gets filled in
  printf("initializing impression equations\n");
  setUpSampleNextSentimentRaw(sstate, BEH_AGENT, RESET_NONE, f, tau);

  printf("testing sampling sentiments\n");
  for (int i=0; i<num_samples; i++) {
    sampleASentiment(sstate, f_out, tau_out);
    theweight = 1.0;
    totalweight += theweight;
    for (int j=0; j<VD; j++) {
      f_avg[j] = f_avg[j]+f_out[j]*theweight;
      tau_avg[j] = tau_avg[j]+tau_out[j]*theweight;
    }
  }
  printf("the total weight: %g\n",totalweight);
  for (int i=0; i<VD; i++) {
    f_avg[i] = f_avg[i]/totalweight;
    tau_avg[i] = tau_avg[i]/totalweight;
  }

  printf("pre-observation fundamental output average (includes optimal behaviour f_b): ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_avg[i]);
  printf("\n");
  printf("pre-observation transient output average (if optimal behaviour taken): ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_avg[i]);
  printf("\n");
  defl = 0.0;
  for (int i=0; i<VD; i++)
    defl += pow(f_avg[i]-tau_avg[i],2);
  printf("deflection (pre-observations bayesact): %g\n",defl);

  // f_avg is the optimal behaviour for this identity set, taking into account only the sentiments
  // compute closest labels
    // behaviour dictionary
  dictionary beh_dictionary;
  int num_behaviours;
  num_behaviours = readSentimentDictionary(MEAN,"fbehaviours.dat",&beh_dictionary);
  printf("num behaviours: %d\n",num_behaviours);
  int index;
  printf("optimal behaviour: \n");
  for (int i=0; i<SSD; i++)
    printf("%g ",f_avg[i+3]);
  printf("\n");
  somaticTransform(DENOTATIVE, sstate->r, dyad_index, &index, &f_avg[3], &beh_dictionary);
  printf("index of optimal: %d\n",index);
  printf("optimal behaviour: %s\n",beh_dictionary.theDictionary[index].label);
  for (int i=0; i<SSD; i++)
    printf("%g ",gsl_vector_get(beh_dictionary.theDictionary[index].epa,i));
  printf("\n");

  dictionary id_dictionary;
  int num_identities;
  num_identities = readSentimentDictionary(MEAN,"fidentities.dat",&id_dictionary);
  printf("num behaviours: %d\n",num_identities);
  printf("optimal identity (agent): \n");
  for (int i=0; i<SSD; i++)
    printf("%g ",f_avg[i]);
  printf("\n");
  somaticTransform(DENOTATIVE, sstate->r, dyad_index, &index, f_avg, &id_dictionary);
  printf("index of optimal identity (agent): %d\n",index);
  printf("optimal re-identification: %s\n",id_dictionary.theDictionary[index].label);
  for (int i=0; i<SSD; i++)
    printf("%g ",gsl_vector_get(id_dictionary.theDictionary[index].epa,i));
  printf("\n");

  
  // now take the observation of behaviour into account and compute the posterior
  for (int i=0; i<VD; i++) {
    f_avg[i] = 0.0;
    tau_avg[i] = 0.0;
  }
  totalweight=0.0;
  for (int i=0; i<num_samples; i++) {
    sampleASentiment(sstate, f_out, tau_out);
    // now compute the weight using exp(-(behaviour_observation - f_out[3:6])^2/gamma^2)
    theweight = 1.0;
    for (int j=0; j<3; j++) {
      theweight *= exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma));
    }
    //theweight = 1.0;
    totalweight += theweight;
    /*
    if (theweight > 0.0) {
      for (int j=0; j<3; j++) 
	printf("%g %g %g\n",f_out[j+3],behaviour_observation[j],exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma)));
      printf("the weight: %g\n",theweight);
    }
    */
    for (int j=0; j<VD; j++) {
      f_avg[j] = f_avg[j]+f_out[j]*theweight;
      tau_avg[j] = tau_avg[j]+tau_out[j]*theweight;
    }
  }
  
  printf("the total weight: %g\n",totalweight);
  if (totalweight > WEIGHT_THRESHOLD) {
    // sufficient weight found in samples, exit function (print at bottom)
    for (int i=0; i<VD; i++) {
      f_avg[i] = f_avg[i]/totalweight;
      tau_avg[i] = tau_avg[i]/totalweight;
    }
    
  } else {
    //*************************************************************************************
    // now we will set up to reset
    // use f_interact so f_b is "shoehorned in" -- since the sig_f value is really small when resetting,
    // this means the behaviour is "set" in some sense as it will be forced to agree with the observation
    // then all the weight for the resetting is put on the client id to be reset
    printf("***************** Weight was zero (%g), so resetting client\n",totalweight);

    computeInteractMean(sstate, BEH_AGENT, f_interact,tau_interact);
    
    // only do the client
    for (int i=0; i<VD; i++) {
      if (i<SSD*2) {
	f_reset[i]=f[i];
      } else {
	f_reset[i] =gsl_vector_get(sstate->sentiParams->interact_mean,i);
      }
    }
    //for (int i=0; i<SSD; i++) 
    //f_reset[i] =gsl_vector_get(sstate->sentiParams[dyad_index].interact_mean,i);
    //for (int i=SSD*2; i<SSD*3; i++) 
    //      f_reset[i] =gsl_vector_get(sstate->sentiParams[dyad_index].interact_mean,i);
    
    for (int i=0; i<VD; i++) 
      printf("f_reset[%d]=%4.3g\n",i,f_reset[i]);
    memcpy(tau_reset,f_reset,VD*sizeof(double));
    
    setUpSampleNextSentimentRaw(sstate, BEH_AGENT, RESET_NONE, f_reset, tau_reset);
    
    // now take the observation of behaviour into account and compute the posterior
    for (int i=0; i<VD; i++) {
      f_avg[i] = 0.0;
      tau_avg[i] = 0.0;
    }
    totalweight=0.0;
    for (int i=0; i<num_samples; i++) {
      sampleASentiment(sstate, f_out, tau_out);
      
      // now compute the weight using exp(-(behaviour_observation - f_out[3:6])^2/gamma^2)
      theweight = 1.0;
      for (int j=0; j<3; j++) {
	theweight *= exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma));
      }
      //theweight = 1.0;
      totalweight += theweight;
      /*
	if (theweight > 0.0) {
	for (int j=0; j<3; j++) 
	printf("%g %g %g\n",f_out[j+3],behaviour_observation[j],exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma)));
	printf("the weight: %g\n",theweight);
	}




      */
      for (int j=0; j<VD; j++) {
	f_avg[j] = f_avg[j]+f_out[j]*theweight;
	tau_avg[j] = tau_avg[j]+tau_out[j]*theweight;
      }
    }
    if (totalweight > WEIGHT_THRESHOLD) {
      // sufficient weight found in samples, exit function (print at bottom)
      printf("the total weight: %g\n",totalweight);
      for (int i=0; i<VD; i++) {
	f_avg[i] = f_avg[i]/totalweight;
	tau_avg[i] = tau_avg[i]/totalweight;
      }
    } else {
      //*************************************************************************************
      // now we will set up to reset
      // resample both identities
      // now we also shoehorn in the behaviour observation in b
      // this is ensured by calling setUpSampleNextSentimentRaw(...RESET_X)   _X to be removed but now is _BOTH
      // 
      printf("***************** Weight was zero (%g), so resetting both agent and client\n",totalweight);
      for (int i=0; i<VD; i++) {
	if (i<SSD || i>=SSD*2) {
	  f_reset[i]=gsl_vector_get(sstate->sentiParams->interact_mean,i);
	} else{
	  f_reset[i]= f_interact[i];
	}
      }
      //for (int i=0; i<SSD; i++) 
      //f_reset[i] =gsl_vector_get(sstate->sentiParams.interact_mean,i);
      //for (int i=SSD*2; i<SSD*3; i++) 
      //      f_reset[i] =gsl_vector_get(sstate->sentiParams.interact_mean,i);
      
      for (int i=0; i<VD; i++) 
	printf("f_reset[%d]=%4.3g\n",i,f_reset[i]);
      memcpy(tau_reset,f_reset,VD*sizeof(double));
      
      setUpSampleNextSentimentRaw(sstate, BEH_AGENT, RESET_BOTH, f_reset, tau_reset);
      
      // now take the observation of behaviour into account and compute the posterior
      for (int i=0; i<VD; i++) {
	f_avg[i] = 0.0;
	tau_avg[i] = 0.0;
      }
      totalweight=0.0;
      for (int i=0; i<num_samples; i++) {
	sampleASentiment(sstate, f_out, tau_out);
	
	// now compute the weight using exp(-(behaviour_observation - f_out[3:6])^2/gamma^2)
	theweight = 1.0;
	for (int j=0; j<3; j++) {
	  theweight *= exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma));
	}
	//theweight = 1.0;
	totalweight += theweight;
	/*
	  if (theweight > 0.0) {
	  for (int j=0; j<3; j++) 
	  printf("%g %g %g\n",f_out[j+3],behaviour_observation[j],exp(-1.0*pow(behaviour_observation[j]-f_out[j+3],2)/(gamma*gamma)));
	  printf("the weight: %g\n",theweight);
	  }
	*/
	for (int j=0; j<VD; j++) {
	  f_avg[j] = f_avg[j]+f_out[j]*theweight;
	  tau_avg[j] = tau_avg[j]+tau_out[j]*theweight;
	}
      }
      
      if (totalweight > WEIGHT_THRESHOLD) {
	printf("the total weight: %g\n",totalweight);
	for (int i=0; i<VD; i++) {
	  f_avg[i] = f_avg[i]/totalweight;
	  tau_avg[i] = tau_avg[i]/totalweight;
	}
      } else {
	// resample from scratch somehow or fail
	printf("failed to generate appropriate samples with weight %g\n",totalweight);
	exit(0);
      }
    }
  }
  /*
  diff = 0.0;
  //double interact_avg[9] = {0.90,-0.09,0.02,0.62,0.07,0.13,0.89,1.17,0.93};
  for (int i=0; i<VD; i++) {
    diff += pow(tau_avg[i]-tau_out_interact[i],2);
  }
  printf("gamma: %g  difference with interact: %g\n",gamma,sqrt(diff));
  printf("%g %g %g %g\n",alpha,beta,gamma,sqrt(diff));
  */

  printf("fundamental output average: ");
  for (int i=0; i<VD; i++)
    printf("%g ",f_avg[i]);
  printf("\n");
  printf("transient output average: ");
  for (int i=0; i<VD; i++)
    printf("%g ",tau_avg[i]);
  printf("\n");
  defl = 0.0;
  for (int i=0; i<VD; i++)
    defl += pow(f_avg[i]-tau_avg[i],2);
  printf("deflection (bayesact): %g\n",defl);

  
  freeParametersFromIndex(sstate);

}
