/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- for Prisoner's dilemma
# which doesn't actually use the X observations but its included here for completeness
#************************************************************************************************/

#include "ultimatumAPP.h"

// initialize and free
void initializeXXObs(xostate **xobs)
{
  *xobs = (xostate *) malloc(sizeof(xostate));
  (*xobs)->obs_amount_won = 0;
}
void freeXXObs(xostate **xobs)
{
  free(*xobs);
}
// update any additional xxstate based on xxstate
// and on the actions taken by the two agents
// which are the indices of the denotative actions taken by agent and client
// in the dictionaries specified as BEHAVIOUR in the input file
// in general this is a draw from P(x'|x'_{ba},x'_{bc})
void updateXXSample(gsl_rng *r, xxstate *next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour)
{

  // mappings are hard-coded -- this needs to be fixed 
  // agent: 
  // X = giveX
  // client - should be: 
  // 0 = reject, 1 = accept
  // but since I can't do this yet (have different sets for agent and client), I just hacked it
  // so that >=5 is "accept" and <5 is "reject" -- see also hack below
  if (client_behaviour<5)  // client rejects - always get zero
      next_xxstate->amount_won = 0;
  else 
    next_xxstate->amount_won = 10-agent_behaviour;
  
  return;
}

// this is Pr(x_{bc}'|x,x_{bc})
void updateXbcSample(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, int num_behaviours)
{
  // client_behaviour is actually an observation but by default its exactly the same
  if (o_bc >= 0) {
    // cheat
    (*x_bc)=o_bc;
  } else {
    // here is a version that choose randomly - this is normally what we should do but it wastes samples
    if (gsl_rng_uniform(r) > 0.5)
      (*x_bc) = 5;   // hack - this should be 1 and 0 
    else 
      (*x_bc) = 0;
  }
}
// this is Pr(x_{ba}'|x,x_{ba})
void updateXbaSample(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, int num_behaviours)
{
  // agent_behaviour is the same as x_ba
  if (a >= 0) {
    // cheat
    (*x_ba)=a;
  } else {
    // here is a version that choose randomly - this is normally what we should do but it wastes samples
    (*x_ba) = gsl_rng_uniform_int(r,11);
  }
}

// this is R(x)
double getXXUtility(xxstate * xx_state)
{
  return xx_state->amount_won;
}

// weight x' based on o_x'
// generally this returns P(o_x | x')
double weightXXSample(gsl_rng * r, xxstate * xx_state, xostate * xx_observation)
{
  if (xx_state->amount_won == xx_observation->obs_amount_won)
    return 1.0-PERCEPTION_NOISE;

  return PERCEPTION_NOISE;
}
// weight x_{bc}' based on \omega_{x_{bc}} - this is P(o'_bc | x'_bc)
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc)
{
  // could be somewhat uncertain about what this agent perceives
  if ((*x_bc)==o_bc)
    return 1-PERCEPTION_NOISE;
  return PERCEPTION_NOISE;
}
// draw a sample from P(o_x | x) 
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation)
{
  // could add noise here
  (*xx_observation).obs_amount_won = xx_state->amount_won;
}
// draw a sample from P(o_bc | x_bc) 
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc)
{
  // could add noise here
  (*o_bc) = x_bc;
}

// other helper functions
int stringToXXObs(char * str, xostate *xx_observation)
{
  xx_observation->obs_amount_won = atoi(str);
  return 1;  // success  -1 for failure
}
void xxObsToString(char * retstr, xostate *xx_observation)
{
  sprintf(retstr,"%d",xx_observation->obs_amount_won);
}

void copyXXstate(xxstate *dest, xxstate *src)
{
  dest->amount_won = src->amount_won;
}

void printRawXXstate(xxstate * xx) {
  printf("amount won: %d\n",xx->amount_won);
}
void printRawXXObs(xostate * xx_observation)
{
  printf("amount won observed %d\n",xx_observation->obs_amount_won);
}
int compareXXObs(xostate * xx1, xostate *xx2)
{
  return xx1->obs_amount_won == xx2->obs_amount_won;
}
void initializeXXstate(xxstate *xx)
{
  xx->amount_won=0;
}
void addToXXstate(xxstate *xx1, xxstate *xx2, double weight)
{
  xx1->amount_won += xx2->amount_won*weight;
}
void divideXXstate(xxstate * xx,double v)
{
  xx->amount_won  = (int) ((double) xx->amount_won)/v;
}


