/************************************************************************************************#
#                           Bayesian Affect Control Theory (V2)                                  #
#************************************************************************************************#
#  Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey                 #
#  Date  : March 2017                                                                            #
#  Copyright (C) 2017  Jesse Hoey                                                                #
#                                                                                                #
#  Use for research purposes only.                                                               #
#  Do not re-distribute without written permission from the author                               #
#  Any commerical uses strictly forbidden.                                                       #
#  This program is distributed WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                                          #
#                                                                                                #
#  See README for details                                                                        #
#  See bayesact.ca for details and instructional videos                                          # 
#************************************************************************************************#
#  application dependent definition of the additional x state -- for Prisoner's dilemma
# which doesn't actually use the X observations but its included here for completeness
#************************************************************************************************/

// application dependent definition of the state
#ifndef __ULTIMATUMAPP
#define __ULTIMATUMAPP

#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_rng.h>


// amount of noise in perceptions of both client behaviour and state (amount won)
#define PERCEPTION_NOISE 0.0

// additional state needs to be specified in a different header file that
// is application specific - we put it here for now
typedef struct xx_state xxstate;
struct xx_state {
  int amount_won;
};

typedef struct xo_state xostate;
struct xo_state {
  int obs_amount_won;
};

// initialize and free
void initializeXXObs(xostate **xobs);
void freeXXObs(xostate **xobs);
// update the x state 
void updateXXSample(gsl_rng * r, xxstate * next_xxstate, xxstate * current_xxstate, int agent_behaviour, int client_behaviour);
void updateXbcSample(gsl_rng *r, int * x_bc,  xxstate current_xxstate, int * current_xstate, int o_bc, int num_behaviours);
void updateXbaSample(gsl_rng *r, int * x_ba,  xxstate current_xxstate, int * current_xstate, int a, int num_behaviours);
double weightXXSample(gsl_rng * r, xxstate * xx_state, xostate * xx_observation);
double weightXXbcSample(gsl_rng * r, int * x_bc, int o_bc);
double getXXUtility(xxstate * xx_state);
void sampleXbc(gsl_rng * r, int x_bc, int *o_bc);
void sampleX(gsl_rng * r, xxstate * xx_state, xostate *xx_observation);
int compareXXObs(xostate * xx1, xostate *xx2);

int stringToXXObs(char * str, xostate *xx_observation);
void xxObsToString(char *retstr, xostate *xx_observation);
void copyXXstate(xxstate *dest, xxstate *src);
void printRawXXstate(xxstate * xx);
void printRawXXObs(xostate * xx_observation);
void initializeXXstate(xxstate *xx);
void addToXXstate(xxstate *xx_avg, xxstate *xx, double weight);
void divideXXstate(xxstate * xx,double v);

#endif
